from datetime import datetime
from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, NodeDTO, PrimaryID


class MetricDTO(BaseDTO):
    metric_id: str = PrimaryID(None)

    key: str = Field(None)
    value: float = Field(None)

    timestamp: Optional[datetime] = Field(None)

    sender: NodeDTO = Field(NodeDTO())

    model_id: str = Field(None)
    step: Optional[int] = Field(None)

    round_id: Optional[str] = Field(None)
    session_id: Optional[str] = Field(None)


"""
log_metric(loss: 1, accarcy: 0.9) # 0
log_metric(loss: 0.9) # 1
log_metric(accarcy: 0.9) # 2
log_metric(loss: 0.8, accarcy: 0.9) # 3

new round

log_metric(loss: 0.8, accarcy: 0.9) # 0"
"""



from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, PrimaryID, validator
from fedn.network.storage.statestore.stores.shared import ValidationError


def allowed_file_extension(filename: str, ALLOWED_EXTENSIONS={"gz", "bz2", "tar", "zip", "tgz"}) -> bool:
    """Check if file extension is allowed.

    :param filename: The filename to check.
    :type filename: str
    :return: True and extension str if file extension is allowed, else False and None.
    :rtype: Tuple (bool, str)
    """
    if "." in filename:
        extension = filename.rsplit(".", 1)[1].lower()
        if extension in ALLOWED_EXTENSIONS:
            return True

    return False


def validate_helper(helper: str) -> bool:
    if not helper or helper == "" or helper not in ["numpyhelper", "binaryhelper", "androidhelper"]:
        return False
    return True


class PackageDTO(BaseDTO):
    """Package data transfer object."""

    package_id: Optional[str] = PrimaryID(None)
    description: str = Field(None)
    file_name: str = Field(None)
    helper: str = Field(None)
    name: str = Field(None)
    storage_file_name: Optional[str] = Field(None)
    active: Optional[bool] = Field(False)

    @validator
    def validate(self):
        if not self.file_name:
            raise ValidationError("file_name", "File name is required")

        if not allowed_file_extension(self.file_name):
            return ValidationError("file_name", "File extension not allowed")

        if not self.helper or not validate_helper(self.helper):
            return ValidationError("helper", "Helper is required or is invalid")



from typing import Optional

from fedn.network.storage.statestore.stores.dto.session import SessionConfigDTO
from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, DictDTO, Field, ListDTO, PrimaryID


class RoundConfigDTO(SessionConfigDTO):
    """RoundConfig data transfer object."""

    session_id: Optional[str] = Field(None)
    task: str = Field(None)
    rounds: int = Field(None)
    round_id: str = Field(None)
    client_settings: Optional[dict] = Field({})


class RoundReduceDTO(DictDTO):
    """RoundReduce data transfer object."""

    time_aggregate_model: Optional[float] = Field(None)
    time_fetch_model: Optional[float] = Field(None)
    time_load_model: Optional[float] = Field(None)


class RoundDataDTO(DictDTO):
    """RoundData data transfer object."""

    time_commit: Optional[float] = Field(None)
    reduce: Optional[RoundReduceDTO] = Field(None)


class AggregationTimeDTO(DictDTO):
    """AggregationTime data transfer object."""

    nr_aggregated_models: Optional[int] = Field(None)
    time_model_aggregation: Optional[float] = Field(None)
    time_model_load: Optional[float] = Field(None)


class RoundCombinerDataDTO(DictDTO):
    """RoundCombinerData data transfer object."""

    aggregation_time: Optional[AggregationTimeDTO] = Field(None)
    nr_expected_updates: Optional[int] = Field(None)
    nr_required_updates: Optional[int] = Field(None)
    time_combination: Optional[float] = Field(None)
    timeout: Optional[float] = Field(None)


class RoundCombinerConfigDTO(RoundConfigDTO):
    """RoundCombinerConfig data transfer object."""

    _job_id: Optional[str] = Field(None)
    round_id: Optional[str] = Field(None)


class RoundCombinerDTO(DictDTO):
    """RoundCombiner data transfer object."""

    round_id: Optional[str] = Field(None)
    model_id: Optional[str] = Field(None)
    name: str = Field(None)
    status: str = Field(None)
    time_exec_training: float = Field(None)
    config: RoundCombinerConfigDTO = Field(None)
    data: RoundCombinerDataDTO = Field(None)


class RoundDTO(BaseDTO):
    """Round data transfer object."""

    round_id: Optional[str] = PrimaryID(None)
    status: Optional[str] = Field(None)
    round_config: Optional[RoundConfigDTO] = Field(None)
    round_data: Optional[RoundDataDTO] = Field(None)
    combiners: ListDTO[RoundCombinerDTO] = Field(ListDTO(RoundCombinerDTO))



import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto.validation import ValidationDTO

@pytest.fixture
def test_validations():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
   
    validation1 = ValidationDTO(validation_id=str(uuid.uuid4()), data="test_data1", timestamp=start_time+datetime.timedelta(days=2), correlation_id=str(uuid.uuid4()))
    validation2 = ValidationDTO(validation_id=str(uuid.uuid4()), data="test_data2", timestamp=start_time+datetime.timedelta(days=4), correlation_id=str(uuid.uuid4()))
    validation3 = ValidationDTO(validation_id=str(uuid.uuid4()), data="test_data3", timestamp=start_time+datetime.timedelta(days=5), correlation_id=str(uuid.uuid4()))
    validation4 = ValidationDTO(validation_id=str(uuid.uuid4()), data="test_data4", timestamp=start_time+datetime.timedelta(days=1), correlation_id=str(uuid.uuid4()))
    validation5 = ValidationDTO(validation_id=str(uuid.uuid4()), data="test_data5", timestamp=start_time+datetime.timedelta(days=6), correlation_id=str(uuid.uuid4()))

    return [validation1, validation2, validation3, validation4, validation5]

 

@pytest.fixture
def test_validation():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
    return ValidationDTO(data="test_data1", timestamp=start_time, correlation_id=str(uuid.uuid4()))



@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_validations):
    for c in test_validations:
        mongo_connection.validation_store.add(c)
        postgres_connection.validation_store.add(c)
        sql_connection.validation_store.add(c)
    
    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    
    for c in test_validations:
        mongo_connection.validation_store.delete(c.validation_id)
        postgres_connection.validation_store.delete(c.validation_id)
        sql_connection.validation_store.delete(c.validation_id)



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "data",
                    "timestamp",
                    "invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"data":"test_data6"}, {"data":""})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestValidationStore:

    def test_add_update_delete(self, db_connection:DatabaseConnection, test_validation:ValidationDTO):
        test_validation.check_validity()

        read_validation1 = db_connection.validation_store.add(test_validation)
        assert isinstance(read_validation1.validation_id, str)
        assert isinstance(read_validation1.committed_at, datetime.datetime)

        read_validation1_dict = read_validation1.to_dict()
        validation_id = read_validation1_dict["validation_id"]
        del read_validation1_dict["validation_id"]
        del read_validation1_dict["committed_at"]

        test_validation_dict = test_validation.to_dict()
        del test_validation_dict["validation_id"]
        del test_validation_dict["committed_at"]

        assert read_validation1_dict == test_validation_dict

        # Assert we get the same validation back
        read_validation2 = db_connection.validation_store.get(validation_id)
        assert read_validation2 is not None
        assert read_validation2.to_dict() == read_validation1.to_dict()    

        # Delete the validation and check that it is deleted
        success = db_connection.validation_store.delete(validation_id)
        assert success == True

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_validations = db_1.validation_store.list(*opt, **kwargs)
                count = db_1.validation_store.count(**kwargs)
                
                gathered_validations2 = db_2.validation_store.list(*opt, **kwargs)
                count2 = db_2.validation_store.count(**kwargs)
                
                assert count == count2
                assert len(gathered_validations) == len(gathered_validations2)

                for i in range(len(gathered_validations)):
                    assert gathered_validations[i].validation_id == gathered_validations2[i].validation_id






import math
import time
import traceback
from typing import Any, Dict, Optional, Tuple

from fedn.common.exceptions import InvalidParameterError
from fedn.common.log_config import logger
from fedn.network.combiner.aggregators.aggregatorbase import AggregatorBase
from fedn.utils.helpers.helperbase import HelperBase
from fedn.utils.parameters import Parameters


class Aggregator(AggregatorBase):
    """Federated Optimization (FedOpt) aggregator.

    Implmentation following: https://arxiv.org/pdf/2003.00295.pdf

    This aggregator computes pseudo gradients by subtracting the model
    update from the global model weights from the previous round.
    A server-side scheme is then applied, currenty supported schemes
    are "adam", "yogi", "adagrad".

    Limitations:
        - Only supports one combiner.
        - Momentum is reser for each new invokation of a training session.

    :param control: A handle to the :class: `fedn.network.combiner.updatehandler.UpdateHandler`
    :type control: class: `fedn.network.combiner.updatehandler.UpdateHandler`

    """

    def __init__(self, update_handler):
        super().__init__(update_handler)

        self.name = "fedopt"
        # To store momentum
        self.v = None
        self.m = None

    def combine_models(
        self, helper: Optional[HelperBase] = None, delete_models: bool = True, parameters: Optional[Parameters] = None
    ) -> Tuple[Optional[Any], Dict[str, float]]:
        """Compute pseudo gradients using model updates in the queue.

        :param helper: ML framework-specific helper, defaults to None.
        :param delete_models: Delete models from storage after aggregation, defaults to True.
        :param parameters: Aggregator hyperparameters, defaults to None.
        :return: The global model and metadata.
        """
        data = {"time_model_load": 0.0, "time_model_aggregation": 0.0}

        # Default hyperparameters
        default_parameters = {
            "serveropt": "adam",
            "learning_rate": 1e-3,
            "beta1": 0.9,
            "beta2": 0.99,
            "tau": 1e-4,
        }

        # Validate and merge parameters
        try:
            parameters = self._validate_and_merge_parameters(parameters, default_parameters)
        except InvalidParameterError as e:
            logger.error(f"Aggregator {self.name} received invalid parameters: {e}")
            return None, data

        logger.info(f"Aggregator {self.name} starting model aggregation.")

        # Aggregation initialization
        pseudo_gradient, model_old = None, None
        nr_aggregated_models, total_examples = 0, 0

        while not self.update_handler.model_updates.empty():
            try:
                logger.info(f"Aggregator {self.name}: Fetching next model update.")
                model_update = self.update_handler.next_model_update()

                tic = time.time()
                model_next, metadata = self.update_handler.load_model_update(model_update, helper)
                data["time_model_load"] += time.time() - tic

                logger.info(f"Processing model update {model_update.model_update_id}")

                # Increment total examples
                total_examples += metadata["num_examples"]

                tic = time.time()
                if nr_aggregated_models == 0:
                    model_old = self.update_handler.load_model(helper, model_update.model_id)
                    pseudo_gradient = helper.subtract(model_next, model_old)
                else:
                    pseudo_gradient_next = helper.subtract(model_next, model_old)
                    pseudo_gradient = helper.increment_average(pseudo_gradient, pseudo_gradient_next, metadata["num_examples"], total_examples)

                data["time_model_aggregation"] += time.time() - tic

                nr_aggregated_models += 1

                if delete_models:
                    self.update_handler.delete_model(model_update)
                    logger.info(f"Deleted model update {model_update.model_update_id} from storage.")
            except Exception as e:
                logger.error(f"Error processing model update: {e}. Skipping this update.")
                logger.error(traceback.format_exc())
                continue

        data["nr_aggregated_models"] = nr_aggregated_models

        if pseudo_gradient:
            try:
                model = self._apply_server_optimizer(helper, pseudo_gradient, model_old, parameters)
            except Exception as e:
                logger.error(f"Error during model aggregation: {e}")
                logger.error(traceback.format_exc())
                return None, data
        else:
            return None, data

        logger.info(f"Aggregator {self.name} completed. Aggregated {nr_aggregated_models} models.")
        return model, data

    def _validate_and_merge_parameters(self, parameters: Optional[Parameters], default_parameters: Dict[str, Any]) -> Dict[str, Any]:
        """Validate and merge default parameters."""
        parameter_schema = {
            "serveropt": str,
            "learning_rate": float,
            "beta1": float,
            "beta2": float,
            "tau": float,
        }
        if parameters:
            parameters.validate(parameter_schema)
        else:
            logger.info(f"Aggregator {self.name} using default parameters.")
            parameters = {}
        return {**default_parameters, **parameters}

    def _apply_server_optimizer(self, helper: HelperBase, pseudo_gradient: Any, model_old: Any, parameters: Dict[str, Any]) -> Any:  # noqa: ANN401
        """Apply the selected server optimizer to compute the new model."""
        optimizer_map = {
            "adam": self.serveropt_adam,
            "yogi": self.serveropt_yogi,
            "adagrad": self.serveropt_adagrad,
        }
        optimizer = optimizer_map.get(parameters["serveropt"])
        if not optimizer:
            raise ValueError(f"Unsupported server optimizer: {parameters['serveropt']}")
        return optimizer(helper, pseudo_gradient, model_old, parameters)

    def serveropt_adam(self, helper, pseudo_gradient, model_old, parameters):
        """Server side optimization, FedAdam.

        :param helper: instance of helper class.
        :type helper: Helper
        :param pseudo_gradient: The pseudo gradient.
        :type pseudo_gradient: As defined by helper.
        :param model_old: The current global model.
        :type model_old: As defined in helper.
        :param parameters: Hyperparamters for the aggregator.
        :type parameters: dict
        :return: new model weights.
        :rtype: as defined by helper.
        """
        beta1 = parameters["beta1"]
        beta2 = parameters["beta2"]
        learning_rate = parameters["learning_rate"]
        tau = parameters["tau"]

        if not self.v:
            self.v = helper.ones(pseudo_gradient, math.pow(tau, 2))

        if not self.m:
            self.m = helper.multiply(pseudo_gradient, [(1.0 - beta1)] * len(pseudo_gradient))
        else:
            self.m = helper.add(self.m, pseudo_gradient, beta1, (1.0 - beta1))

        p = helper.power(pseudo_gradient, 2)
        self.v = helper.add(self.v, p, beta2, (1.0 - beta2))

        sv = helper.add(helper.sqrt(self.v), helper.ones(self.v, tau))
        t = helper.divide(self.m, sv)
        model = helper.add(model_old, t, 1.0, learning_rate)

        return model

    def serveropt_yogi(self, helper, pseudo_gradient, model_old, parameters):
        """Server side optimization, FedYogi.

        :param helper: instance of helper class.
        :type helper: Helper
        :param pseudo_gradient: The pseudo gradient.
        :type pseudo_gradient: As defined by helper.
        :param model_old: The current global model.
        :type model_old: As defined in helper.
        :param parameters: Hyperparamters for the aggregator.
        :type parameters: dict
        :return: new model weights.
        :rtype: as defined by helper.
        """
        beta1 = parameters["beta1"]
        beta2 = parameters["beta2"]
        learning_rate = parameters["learning_rate"]
        tau = parameters["tau"]

        if not self.v:
            self.v = helper.ones(pseudo_gradient, math.pow(tau, 2))

        if not self.m:
            self.m = helper.multiply(pseudo_gradient, [(1.0 - beta1)] * len(pseudo_gradient))
        else:
            self.m = helper.add(self.m, pseudo_gradient, beta1, (1.0 - beta1))

        p = helper.power(pseudo_gradient, 2)
        s = helper.sign(helper.add(self.v, p, 1.0, -1.0))
        s = helper.multiply(s, p)
        self.v = helper.add(self.v, s, 1.0, -(1.0 - beta2))

        sv = helper.add(helper.sqrt(self.v), helper.ones(self.v, tau))
        t = helper.divide(self.m, sv)
        model = helper.add(model_old, t, 1.0, learning_rate)

        return model

    def serveropt_adagrad(self, helper, pseudo_gradient, model_old, parameters):
        """Server side optimization, FedAdam.

        :param helper: instance of helper class.
        :type helper: Helper
        :param pseudo_gradient: The pseudo gradient.
        :type pseudo_gradient: As defined by helper.
        :param model_old: The current global model.
        :type model_old: As defined in helper.
        :param parameters: Hyperparamters for the aggregator.
        :type parameters: dict
        :return: new model weights.
        :rtype: as defined by helper.
        """
        beta1 = parameters["beta1"]
        learning_rate = parameters["learning_rate"]
        tau = parameters["tau"]

        if not self.v:
            self.v = helper.ones(pseudo_gradient, math.pow(tau, 2))

        if not self.m:
            self.m = helper.multiply(pseudo_gradient, [(1.0 - beta1)] * len(pseudo_gradient))
        else:
            self.m = helper.add(self.m, pseudo_gradient, beta1, (1.0 - beta1))

        p = helper.power(pseudo_gradient, 2)
        self.v = helper.add(self.v, p, 1.0, 1.0)

        sv = helper.add(helper.sqrt(self.v), helper.ones(self.v, tau))
        t = helper.divide(self.m, sv)
        model = helper.add(model_old, t, 1.0, learning_rate)

        return model



from typing import Tuple
import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto.model import ModelDTO
from fedn.network.storage.statestore.stores.dto.round import RoundDTO



@pytest.fixture
def test_rounds():

    return [RoundDTO(round_id=str(uuid.uuid4()), status="test_status6"),
            RoundDTO(round_id=str(uuid.uuid4()), status="test_status4"),
            RoundDTO(round_id=str(uuid.uuid4()), status="test_status3"),
            RoundDTO(round_id=str(uuid.uuid4()), status="test_status1"),
            RoundDTO(round_id=str(uuid.uuid4()), status="test_status2"),
            RoundDTO(round_id=str(uuid.uuid4()), status="test_status5")]

@pytest.fixture
def test_round_model():
    model = ModelDTO(model_id=str(uuid.uuid4()), name="test_model")

    round = RoundDTO(round_id=str(uuid.uuid4()), status="test_status6")
    round.round_config = {"aggregator":"test_aggregator", "buffer_size":100, "delete_models_storage":True, "clients_required":10,
                          "requested_clients":10,"helper_type":"test_helper_type", "model_id":model.model_id, "round_timeout":100,
                          "validate":True, "round_id":round.round_id, "task":"test_task", "rounds":1, "client_settings":{"test_key":"test_value"}}
    
    round.round_data = {"time_commit":100}
    round.round_data.reduce = {"time_aggregate_model":100, "time_fetch_model":100, "time_load_model":100}
    round.combiners = [{"model_id":model.model_id, "name":"test_name1", "status":"test_status1", "time_exec_training":100,
                              "config":round.round_config.to_dict(),
                              "data":{"aggregation_time":{"nr_aggregated_models":10, "time_model_aggregation":100, "time_model_load":100},
                                      "nr_expected_updates":10, "nr_required_updates":10, "time_combination":100, "timeout":100}}]
    return round, model


@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_rounds):
    for c in test_rounds:
        mongo_connection.round_store.add(c)
        postgres_connection.round_store.add(c)
        sql_connection.round_store.add(c)

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_rounds:
        mongo_connection.round_store.delete(c.round_id)
        postgres_connection.round_store.delete(c.round_id)
        sql_connection.round_store.delete(c.round_id)


    



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "status", 
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"status":"test_status4,test_status5"}, {"status":"blah"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestRoundStore:

    def test_add_update_delete(self, db_connection: DatabaseConnection, test_round_model: Tuple[RoundDTO, ModelDTO]):
        test_round, test_model = test_round_model

        db_connection.model_store.add(test_model)

        test_round.check_validity()

        # Add a round and check that we get the added round back
        read_round1 = db_connection.round_store.add(test_round)
        assert isinstance(read_round1.round_id, str)
        read_round1_dict = read_round1.to_dict()
        round_id = read_round1_dict["round_id"]
        del read_round1_dict["round_id"]
        del read_round1_dict["committed_at"]

        test_round_dict = test_round.to_dict()
        del test_round_dict["round_id"]
        del test_round_dict["committed_at"]

        assert read_round1_dict == test_round_dict

        # Assert we get the same round back
        read_round2 = db_connection.round_store.get(round_id)
        assert read_round2 is not None
        assert read_round2.to_dict() == read_round1.to_dict()
        
        # Update the round and check that we get the updated round back
        read_round2.round_config.aggregator = "new_aggregator"         
        read_round3 = db_connection.round_store.update(read_round2)
        assert read_round3.round_config.aggregator == "new_aggregator"

        # Assert we get the same round back
        read_round4 = db_connection.round_store.get(round_id)
        assert read_round4 is not None
        assert read_round3.to_dict() == read_round4.to_dict()

        # Add combiner to round and check that we get the updated round back
        combiner = read_round4.combiners[0]
        combiner.status = "new_status"
        read_round4.combiners.append(combiner)
        read_round5 = db_connection.round_store.update(read_round4)
        assert len(read_round5.combiners) == 2
        assert read_round5.combiners[1].status == "new_status"

        # Assert we get the same round back
        read_round6 = db_connection.round_store.get(round_id)
        assert read_round6 is not None
        assert read_round5.to_dict() == read_round6.to_dict()


        # Remove config from round and check that we get the updated round back
        read_round6.round_config = None
        read_round7 = db_connection.round_store.update(read_round6)
        assert read_round7.to_dict()["round_config"] == None

        # Assert we get the same round back
        read_round8 = db_connection.round_store.get(round_id)
        assert read_round8 is not None
        assert read_round7.to_dict() == read_round8.to_dict()

        #Remove all combiners from round and check that we get the updated round back
        read_round8.combiners = []
        read_round9 = db_connection.round_store.update(read_round8)
        assert len(read_round9.combiners) == 0

        # Assert we get the same round back
        read_round10 = db_connection.round_store.get(round_id)
        assert read_round10 is not None
        assert read_round9.to_dict() == read_round10.to_dict()
        

        # Delete the round and check that it is deleted
        success = db_connection.round_store.delete(round_id)
        assert success == True
        
        db_connection.model_store.delete(test_model.model_id)


    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_rounds = db_1.round_store.list(*opt, **kwargs)
                count = db_1.round_store.count(**kwargs)

                gathered_rounds2 = db_2.round_store.list(*opt, **kwargs)
                count2 = db_2.round_store.count(**kwargs)
                
                assert count == count2
                assert len(gathered_rounds) == len(gathered_rounds2)

                for i in range(len(gathered_rounds)):
                    assert gathered_rounds2[i].round_id == gathered_rounds[i].round_id





from datetime import datetime
from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, NodeDTO, PrimaryID


class ValidationDTO(BaseDTO):
    """Validation data transfer object."""

    # TODO: Correct which fields are optional
    validation_id: Optional[str] = PrimaryID(None)
    model_id: Optional[str] = Field(None)
    data: Optional[str] = Field(None)
    correlation_id: Optional[str] = Field(None)
    timestamp: Optional[datetime] = Field(None)
    session_id: Optional[str] = Field(None)
    meta: Optional[str] = Field(None)
    sender: Optional[NodeDTO] = Field(NodeDTO())
    receiver: Optional[NodeDTO] = Field(NodeDTO())



from datetime import datetime, timedelta, timezone
from typing import Dict, List

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.analytic import AnalyticDTO
from fedn.network.storage.statestore.stores.store import MongoDBStore, Store, from_document


class AnalyticStore(Store[AnalyticDTO]):
    pass


class MongoDBAnalyticStore(AnalyticStore, MongoDBStore[AnalyticDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "id")
        self.database[self.collection].create_index([("sender_id", pymongo.DESCENDING)])

    def add(self, item: AnalyticDTO) -> AnalyticDTO:
        analytic = super().add(item)
        self._delete_old_records(analytic.sender_id)
        return analytic

    def _delete_old_records(self, sender_id: str) -> int:
        time_threshold = datetime.now(timezone.utc) - timedelta(minutes=5)

        result = self.database[self.collection].delete_many({"sender_id": sender_id, "committed_at": {"$lt": time_threshold}})
        return result.deleted_count

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> List[AnalyticDTO]:
        return super().list(limit, skip, sort_key or "committed_at", sort_order, **kwargs)

    def _document_from_dto(self, item: AnalyticDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> AnalyticDTO:
        item = from_document(document)
        return AnalyticDTO().patch_with(item, throw_on_extra_keys=False)



from typing import Dict

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.session import SessionConfigDTO, SessionDTO
from fedn.network.storage.statestore.stores.sql.shared import SessionConfigModel, SessionModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class SessionStore(Store[SessionDTO]):
    pass


class MongoDBSessionStore(SessionStore, MongoDBStore[SessionDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "session_id")
        self.database[self.collection].create_index([("session_id", pymongo.DESCENDING)])

    def update(self, item: SessionDTO) -> SessionDTO:
        return self.mongo_update(item)

    def _document_from_dto(self, item: SessionDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> SessionDTO:
        session = from_document(document)
        session_config_dto = SessionConfigDTO().patch_with(session["session_config"], throw_on_extra_keys=False)
        session["session_config"] = session_config_dto

        return SessionDTO().patch_with(session, throw_on_extra_keys=False)


class SQLSessionStore(SessionStore, SQLStore[SessionDTO, SessionModel]):
    def __init__(self, Session):
        super().__init__(Session, SessionModel)

    def update(self, item: SessionDTO) -> SessionDTO:
        return self.sql_update(item)

    def _update_orm_model_from_dto(self, entity: SessionModel, item: SessionDTO) -> SessionModel:
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("session_id", None)

        session_config_dict = item_dict.pop("session_config")
        if entity.session_config is None:
            entity.session_config = SessionConfigModel(**session_config_dict)
        else:
            for key, value in session_config_dict.items():
                setattr(entity.session_config, key, value)

        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: SessionModel) -> SessionDTO:
        session_dict = from_orm_model(item, SessionModel)
        session_config_dict = from_orm_model(item.session_config, SessionConfigModel)

        session_dict["session_config"] = session_config_dict
        session_dict["session_id"] = session_dict.pop("id")

        session_config_dict.pop("id")
        session_config_dict.pop("committed_at")
        session_dict.pop("session_config_id")

        return SessionDTO().populate_with(session_dict)

    def _to_orm_dict(self, item_dict: Dict) -> Dict:
        item_dict["id"] = item_dict.pop("session_id", None)
        return item_dict



import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto.status import StatusDTO

@pytest.fixture
def test_statuses():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
   
    status1 = StatusDTO(status_id=str(uuid.uuid4()), status="status1", timestamp=start_time+datetime.timedelta(days=2), log_level="test_log_level1", type="test_type", sender={"name":"test_sender", "role":"test_role"})
    status2 = StatusDTO(status_id=str(uuid.uuid4()), status="status2", timestamp=start_time+datetime.timedelta(days=4), log_level="test_log_level2", type="test_type", sender={"name":"test_sender", "role":"test_role"})
    status3 = StatusDTO(status_id=str(uuid.uuid4()), status="status3", timestamp=start_time+datetime.timedelta(days=5), log_level="test_log_level3", type="test_type", sender={"name":"test_sender", "role":"test_role"})
    status4 = StatusDTO(status_id=str(uuid.uuid4()), status="status4", timestamp=start_time+datetime.timedelta(days=1), log_level="test_log_level4", type="test_type2", sender={"name":"test_sender", "role":"test_role"})
    status5 = StatusDTO(status_id=str(uuid.uuid4()), status="status5", timestamp=start_time+datetime.timedelta(days=6), log_level="test_log_level5", type="test_type2", sender={"name":"test_sender", "role":"test_role"})

    return [status1, status2, status3, status4, status5]

@pytest.fixture
def test_status():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
    return StatusDTO(status_id=str(uuid.uuid4()), status="status1", timestamp=start_time, log_level="test_log_level1", type="test_type", sender={"name":"test_sender", "role":"test_role"})

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_statuses):
    for c in test_statuses:
        mongo_connection.status_store.add(c)
        postgres_connection.status_store.add(c)
        sql_connection.status_store.add(c)
        
    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    
    for c in test_statuses:
        mongo_connection.status_store.delete(c.status_id)
        postgres_connection.status_store.delete(c.status_id)
        sql_connection.status_store.delete(c.status_id)


@pytest.fixture
def options():
    sorting_keys = (None, 
                    "log_level",
                    "timestamp",
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"log_level":"test_log_level6"}, {"type":"test_type2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))




class TestStatusStore:

    def test_add_update_delete(self, db_connection: DatabaseConnection, test_status: StatusDTO):
        test_status.check_validity()

        read_status1 = db_connection.status_store.add(test_status)
        assert isinstance(read_status1.status_id, str)
        assert isinstance(read_status1.committed_at, datetime.datetime)

        read_status1_dict = read_status1.to_dict()
        status_id = read_status1_dict["status_id"]
        del read_status1_dict["status_id"]
        del read_status1_dict["committed_at"]

        test_status_dict = test_status.to_dict()
        del test_status_dict["status_id"]
        del test_status_dict["committed_at"]

        assert read_status1_dict == test_status_dict

        # Assert we get the same status back
        read_status2 = db_connection.status_store.get(status_id)
        assert read_status2 is not None
        assert read_status2.to_dict() == read_status1.to_dict()    

        # Delete the status and check that it is deleted
        success = db_connection.status_store.delete(status_id)
        assert success == True

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_statuses = db_1.status_store.list(*opt, **kwargs)
                count =  db_1.status_store.count(**kwargs)

                gathered_statuses2 = db_2.status_store.list(*opt, **kwargs)
                count2 =  db_2.status_store.count(**kwargs)

                assert count == count2
                assert len(gathered_statuses) == len(gathered_statuses2)

                for i in range(len(gathered_statuses)):
                    assert gathered_statuses[i].status_id == gathered_statuses2[i].status_id





import datetime
import itertools
import time
import uuid

import pymongo
import pytest

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto import ModelDTO
from fedn.network.storage.statestore.stores.dto.session import (
    SessionConfigDTO, SessionDTO)


@pytest.fixture
def test_sessions():
    model = ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name1")

    session_config = {"aggregator":"test_aggregator", "round_timeout":100, "buffer_size":100, "delete_models_storage":True, 
                      "clients_required":10, "validate":True, "helper_type":"test_helper_type", "model_id":model.model_id, "rounds": 10}

    session1 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname1",  session_config=SessionConfigDTO().patch_with(session_config))
    session2 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname2",  session_config=SessionConfigDTO().patch_with(session_config))
    session3 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname3",  session_config=SessionConfigDTO().patch_with(session_config))
    session4 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname4",  session_config=SessionConfigDTO().patch_with(session_config))
    session5 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname5",  session_config=SessionConfigDTO().patch_with(session_config))
    session6 = SessionDTO(session_id=str(uuid.uuid4()), name="sessionname6",  session_config=SessionConfigDTO().patch_with(session_config))

    return model,[session1, session2, session3, session4, session5, session6]

@pytest.fixture
def test_session_and_model():
    model = ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name1")

    session_config = {"aggregator":"test_aggregator", "round_timeout":100, "buffer_size":100, "delete_models_storage":True, 
                      "clients_required":10, "validate":True, "helper_type":"test_helper_type", "model_id":model.model_id, "rounds": 10}

    session = SessionDTO(name="sessionname",  session_config=SessionConfigDTO().patch_with(session_config))

    return model, session

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_sessions):
    model, test_sessions = test_sessions
    
    mongo_connection.model_store.add(model)    
    postgres_connection.model_store.add(model)
    sql_connection.model_store.add(model)
    

    for c in test_sessions:
        mongo_connection.session_store.add(c)
        postgres_connection.session_store.add(c)
        sql_connection.session_store.add(c)
        

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_sessions:
        mongo_connection.session_store.delete(c.session_id)
        postgres_connection.session_store.delete(c.session_id)
        sql_connection.session_store.delete(c.session_id)
    
    mongo_connection.model_store.delete(model.model_id)
    postgres_connection.model_store.delete(model.model_id)
    sql_connection.model_store.delete(model.model_id)



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "name",
                    "committed_at",
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"status":"test_status4"}, {"status":"blah"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))


class TestSessionStore:

    def test_add_update_delete(self, db_connection: DatabaseConnection, test_session_and_model: tuple[ModelDTO, SessionDTO]):
        model, session = test_session_and_model
        
        db_connection.model_store.add(model)
        # Add a session and check that we get the added session back
        read_session1 = db_connection.session_store.add(session)
        assert isinstance(read_session1.session_id, str)
        assert isinstance(read_session1.committed_at, datetime.datetime)
        read_session1_dict = read_session1.to_dict()

        assert read_session1_dict["name"] == session.name
        

        session_config_dict = session.session_config.to_dict()
        assert read_session1_dict["session_config"] == session_config_dict

        session_id = read_session1_dict["session_id"]

        # Assert we get the same session back
        read_session2 = db_connection.session_store.get(session_id)
        assert read_session2 is not None
        assert read_session2.to_dict() == read_session1.to_dict()
        
        # Update the session and check that we get the updated session back
        read_session2.name = "new_name"         
        read_session3 = db_connection.session_store.update(read_session2)
        assert read_session3.name == "new_name"

        # Assert we get the same session back
        read_session4 = db_connection.session_store.get(session_id)
        assert read_session4 is not None
        assert read_session3.to_dict() == read_session4.to_dict()

        # Delete the session and check that it is deleted
        success = db_connection.session_store.delete(session_id)
        assert success == True
        db_connection.model_store.delete(model.model_id)



    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                gathered_sessions = db_1.session_store.list(*opt, **kwargs)
                count = db_1.session_store.count(**kwargs)

                gathered_sessions2 = db_2.session_store.list(*opt, **kwargs)
                count2 = db_2.session_store.count(**kwargs)
                
                assert(len(gathered_sessions) == len(gathered_sessions2))
                assert count == count2

                for i in range(len(gathered_sessions)):
                    assert gathered_sessions2[i].session_id == gathered_sessions[i].session_id





import os

from fedn.common.log_config import logger
from fedn.network.combiner.interfaces import CombinerInterface
from fedn.network.loadbalancer.leastpacked import LeastPacked
from fedn.network.storage.statestore.stores.client_store import ClientStore
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore
from fedn.network.storage.statestore.stores.dto import ClientDTO

__all__ = ("Network",)


class Network:
    """FEDn network interface. This class is used to interact with the network.
    Note: This class contain redundant code, which is not used in the current version of FEDn.
    Some methods has been moved to :class:`fedn.network.api.interface.API`.
    """

    def __init__(self, control, network_id: str, combiner_store: CombinerStore, client_store: ClientStore, load_balancer=None):
        """ """
        self.combiner_store = combiner_store
        self.client_store = client_store
        self.control = control
        self.id = network_id

        if not load_balancer:
            self.load_balancer = LeastPacked(self)
        else:
            self.load_balancer = load_balancer

    def get_combiner(self, name):
        """Get combiner by name.

        :param name: name of combiner
        :type name: str
        :return: The combiner instance object
        :rtype: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        """
        combiners = self.get_combiners()
        for combiner in combiners:
            if name == combiner.name:
                return combiner
        return None

    def get_combiners(self):
        """Get all combiners in the network.

        :return: list of combiners objects
        :rtype: list(:class:`fedn.network.combiner.interfaces.CombinerInterface`)
        """
        result = self.combiner_store.list(limit=0, skip=0, sort_key=None)
        combiners = []
        for combiner in result:
            name = combiner.name.upper()
            # General certificate handling, same for all combiners.
            if os.environ.get("FEDN_GRPC_CERT_PATH"):
                with open(os.environ.get("FEDN_GRPC_CERT_PATH"), "rb") as f:
                    cert = f.read()
            # Specific certificate handling for each combiner.
            elif os.environ.get(f"FEDN_GRPC_CERT_PATH_{name}"):
                cert_path = os.environ.get(f"FEDN_GRPC_CERT_PATH_{name}")
                with open(cert_path, "rb") as f:
                    cert = f.read()
            else:
                cert = None
            combiners.append(
                CombinerInterface(combiner.parent, combiner.name, combiner.address, combiner.fqdn, combiner.port, certificate=cert, ip=combiner.ip)
            )

        return combiners

    def find_available_combiner(self):
        """Find an available combiner in the network.

        :return: The combiner instance object
        :rtype: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        """
        combiner = self.load_balancer.find_combiner()
        return combiner

    def handle_unavailable_combiner(self, combiner):
        """This callback is triggered if a combiner is found to be unresponsive.

        :param combiner: The combiner instance object
        :type combiner: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        :return: None
        """
        # TODO: Implement strategy to handle an unavailable combiner.
        logger.warning("REDUCER CONTROL: Combiner {} unavailable.".format(combiner.name))

    # TODO: This method is not used in the current version of FEDn.
    def add_client(self, client: ClientDTO):
        """Add a new client to the network.

        :param client: The client instance object
        :type client: dict
        :return: None
        """
        if self.get_client(client["client_id"]):
            return

        logger.info("adding client {}".format(client.to_dict()))
        self.client_store.add(client)

    # TODO: This method is not used in the current version of FEDn.
    def get_client(self, client_id: str):
        """Get client by client_id.

        :param client_id: client_id of client
        :type client_id: str
        :return: The client instance object
        :rtype: ObjectId
        """
        try:
            client = self.client_store.get(client_id)
            return client
        except Exception:
            return None



# Unittest for Flask API endpoints
#
# Run with:
# python -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#

import unittest
from unittest.mock import patch, MagicMock

from fedn.network.storage.statestore.stores.dto.metric import MetricDTO
import pymongo

from build.lib.fedn.network.storage.statestore.stores.dto.client import ClientDTO
import fedn
from fedn.network.storage.statestore.stores.dto.combiner import CombinerDTO
from fedn.network.storage.statestore.stores.dto.model import ModelDTO
from fedn.network.storage.statestore.stores.dto.package import PackageDTO
from fedn.network.storage.statestore.stores.dto.round import RoundDTO
from fedn.network.storage.statestore.stores.dto.session import SessionDTO
from fedn.network.storage.statestore.stores.dto.status import StatusDTO
from fedn.network.storage.statestore.stores.dto.validation import ValidationDTO  # noqa: F401

entitites = ['clients', 'combiners', 'models', 'packages', 'rounds', 'sessions', 'statuses', 'validations', 'metrics']
keys = ['client_id', 'combiner_id', 'model_id', 'package_id', 'round_id', 'session_id', 'status_id', 'validation_id', 'metric_id']

class NetworkAPITests(unittest.TestCase):
    """ Unittests for the Network API. """
    @patch('fedn.network.controller.controlbase.ControlBase', autospec=True)
    def setUp(self, mock_control):
        # start Flask server in testing mode
        import fedn.network.api.server
        self.app = fedn.network.api.server.app.test_client()


    def test_health(self):
        """ Test get_models endpoint. """
        response = self.app.get('/health')
        # Assert response
        self.assertEqual(response.status_code, 200)

    def test_add_combiner(self):
        """ Test get_models endpoint. """
        # Mock api.get_models
        
        response = self.app.post('/add_combiner')
        
        self.assertEqual(response.status_code, 410)

    def test_get_controller_status(self):
        """ Test get_models endpoint. """
        response = self.app.get('/get_controller_status')
        # Assert response
        self.assertEqual(response.status_code, 200)

    def test_get_endpoints(self):
        """ Test allt get endpoints. """
        excepted_return_count = 1
        expected_return_id = "test"
        fedn.network.api.shared.client_store.list = MagicMock(return_value=[ClientDTO(client_id="test")])
        fedn.network.api.shared.client_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.combiner_store.list = MagicMock(return_value=[CombinerDTO(combiner_id="test")])
        fedn.network.api.shared.combiner_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.model_store.list = MagicMock(return_value=[ModelDTO(model_id="test")])
        fedn.network.api.shared.model_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.package_store.list = MagicMock(return_value=[PackageDTO(package_id="test")])
        fedn.network.api.shared.package_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.round_store.list = MagicMock(return_value=[RoundDTO(round_id="test")])
        fedn.network.api.shared.round_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.session_store.list = MagicMock(return_value=[SessionDTO(session_id="test")])
        fedn.network.api.shared.session_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.status_store.list = MagicMock(return_value=[StatusDTO(status_id="test")])
        fedn.network.api.shared.status_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.validation_store.list = MagicMock(return_value=[ValidationDTO(validation_id="test")])
        fedn.network.api.shared.validation_store.count = MagicMock(return_value=1)
        fedn.network.api.shared.metric_store.list = MagicMock(return_value=[MetricDTO(metric_id="test")])
        fedn.network.api.shared.metric_store.count = MagicMock(return_value=1)


        for key,entity in zip(keys, entitites):
            response = self.app.get(f'/api/v1/{entity}/')
            # Assert response
            self.assertEqual(response.status_code, 200)

            count = response.json['count']

            self.assertEqual(count, excepted_return_count)

            id = response.json['result'][0][key]

            self.assertEqual(id, expected_return_id)

        fedn.network.api.shared.client_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.client_store.list.assert_called_once()
        fedn.network.api.shared.combiner_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.combiner_store.list.assert_called_once()
        fedn.network.api.shared.model_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.model_store.list.assert_called_once()
        fedn.network.api.shared.package_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.package_store.list.assert_called_once()
        fedn.network.api.shared.round_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.round_store.list.assert_called_once()        
        fedn.network.api.shared.session_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.session_store.list.assert_called_once()
        fedn.network.api.shared.status_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.status_store.list.assert_called_once()
        fedn.network.api.shared.validation_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.validation_store.list.assert_called_once()
        fedn.network.api.shared.metric_store.list.assert_called_with(0, 0, None, pymongo.DESCENDING)
        fedn.network.api.shared.metric_store.list.assert_called_once()
   
        for entity in entitites:
            headers = {
                "X-Limit": 10,
                "X-Skip": 10,
                "X-Sort-Key": "test",
                "X-Sort-Order": "asc"
            }
            response = self.app.get(f'/api/v1/{entity}/', headers=headers)
            # Assert response
            self.assertEqual(response.status_code, 200)

        fedn.network.api.shared.client_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.combiner_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.model_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.package_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.round_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)        
        fedn.network.api.shared.session_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.status_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.validation_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
        fedn.network.api.shared.metric_store.list.assert_called_with(10, 10, "test", pymongo.ASCENDING)
           

if __name__ == '__main__':
    unittest.main()



import uuid
from abc import abstractmethod
from datetime import datetime
from typing import Dict, List

import pymongo
from pymongo.database import Database
from sqlalchemy import select
from werkzeug.utils import secure_filename

from fedn.network.storage.statestore.stores.dto import PackageDTO
from fedn.network.storage.statestore.stores.dto.package import validate_helper
from fedn.network.storage.statestore.stores.sql.shared import PackageModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


def is_active_package(package_id, active_package: dict) -> bool:
    if active_package and "package_id" in active_package and package_id:
        return active_package["package_id"] == package_id
    return False


def _complement_with_storage_filename(item: Dict):
    extension = item["file_name"].rsplit(".", 1)[1].lower()

    if "storage_file_name" not in item or item["storage_file_name"] is None:
        storage_file_name = secure_filename(f"{str(uuid.uuid4())}.{extension}")
        item["storage_file_name"] = storage_file_name


class PackageStore(Store[PackageDTO]):
    @abstractmethod
    def set_active(self, id: str) -> bool:
        """Set the active entity
        param id: The id of the entity
            type: str
        return: Whether the operation was successful
        """
        pass

    @abstractmethod
    def get_active(self) -> PackageDTO:
        """Get the active entity
        return: The entity
        """
        pass

    @abstractmethod
    def set_active_helper(self, helper: str) -> bool:
        """Set the active helper
        param helper: The helper to set as active
            type: str
        return: Whether the operation was successful
        """
        pass

    @abstractmethod
    def delete_active(self) -> bool:
        """Delete the active entity
        return: Whether the operation was successful
        """
        pass


class MongoDBPackageStore(PackageStore, MongoDBStore[PackageDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "package_id")

    def get(self, id: str) -> PackageDTO:
        package = super().get(id)
        if package is None:
            return None

        response_active = self.database[self.collection].find_one({"key": "active"})
        active = is_active_package(package.package_id, response_active)
        package.active = active

        return package

    def delete(self, id: str) -> bool:
        kwargs = {self.primary_key: id, "key": "package_trail"}
        result = self.database[self.collection].delete_one(kwargs).deleted_count == 1
        if not result:
            return False

        # Remove Active Package if it is the one being deleted
        kwargs["key"] = "active"
        document_active = self.database[self.collection].find_one(kwargs)
        if document_active is not None:
            return self.database[self.collection].delete_one(kwargs).deleted_count == 1
        return True

    def delete_active(self) -> bool:
        kwargs = {"key": "active"}
        return self.database[self.collection].delete_one(kwargs).deleted_count == 1

    def list(self, limit=0, skip=0, sort_key=None, sort_order=pymongo.DESCENDING, **kwargs) -> List[PackageDTO]:
        kwargs["key"] = "package_trail"
        packages = super().list(limit, skip, sort_key, sort_order, **kwargs)

        if packages is None or len(packages) == 0:
            return []

        active_package = self.get_active()

        if active_package:
            for package in packages:
                package.active = is_active_package(package.package_id, active_package.to_dict())

        return packages

    def count(self, **kwargs) -> int:
        kwargs["key"] = "package_trail"
        return super().count(**kwargs)

    def set_active(self, id: str) -> bool:
        kwargs = {self.primary_key: id}
        kwargs["key"] = "package_trail"

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return False

        # TODO: Use seperate table to store active package
        activePackage = PackageDTO()
        activePackage.patch_with(document, throw_on_extra_keys=False)
        activePackage.active = True
        activePackage.committed_at = datetime.now()
        obj_to_insert = {"key": "active", **activePackage.to_db()}

        self.database[self.collection].update_one({"key": "active"}, {"$set": obj_to_insert}, upsert=True)

        return True

    def get_active(self) -> PackageDTO:
        kwargs = {"key": "active"}
        document = self.database[self.collection].find_one(kwargs)
        if document is None:
            return None
        document["active"] = True
        return self._dto_from_document(document)

    def set_active_helper(self, helper: str) -> bool:
        if not validate_helper(helper):
            raise ValueError
        try:
            self.database[self.collection].update_one({"key": "active"}, {"$set": {"helper": helper}}, upsert=True)
        except Exception:
            return False

    def _document_from_dto(self, item: PackageDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        item_dict["key"] = "package_trail"
        _complement_with_storage_filename(item_dict)
        return item_dict

    def _dto_from_document(self, document: Dict) -> PackageDTO:
        item_dict = from_document(document)
        del item_dict["key"]
        return PackageDTO().patch_with(item_dict, throw_on_extra_keys=False)


class SQLPackageStore(PackageStore, SQLStore[PackageDTO, PackageModel]):
    def __init__(self, Session):
        super().__init__(Session, PackageModel)

    def set_active(self, id: str):
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()

            stmt = select(PackageModel).where(PackageModel.id == id)
            new_active_item = session.scalars(stmt).first()
            if new_active_item is None:
                return False

            new_active_item.active = True
            if active_item:
                active_item.active = False
            session.commit()
        return True

    def get_active(self) -> PackageDTO:
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                return self._dto_from_orm_model(active_item)
            return None

    def set_active_helper(self, helper: str) -> bool:
        if not validate_helper(helper):
            raise ValueError()

        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.helper = helper
                session.commit()
                return True
            item = PackageModel(
                committed_at=datetime.now(),
                description="",
                file_name="",
                helper=helper,
                name="",
                storage_file_name="",
                active=True,
            )

            session.add(item)
            session.commit()

    def delete_active(self) -> bool:
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.active = False
                session.commit()
                return True
            return False

    def _update_orm_model_from_dto(self, entity: PackageModel, item: PackageDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("package_id", None)
        _complement_with_storage_filename(item_dict)
        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: PackageModel) -> PackageDTO:
        orm_dict = from_orm_model(item, PackageModel)
        orm_dict["package_id"] = orm_dict.pop("id")
        return PackageDTO().populate_with(orm_dict)



from abc import abstractmethod
from typing import Dict

from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto import CombinerDTO
from fedn.network.storage.statestore.stores.sql.shared import CombinerModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class CombinerStore(Store[CombinerDTO]):
    @abstractmethod
    def get_by_name(name: str) -> CombinerDTO:
        pass


class MongoDBCombinerStore(CombinerStore, MongoDBStore[CombinerDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "combiner_id")

    def get_by_name(self, name: str) -> CombinerDTO:
        document = self.database[self.collection].find_one({"name": name})
        if document is None:
            return None
        return self._dto_from_document(document)

    def _document_from_dto(self, item: CombinerDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> CombinerDTO:
        return CombinerDTO().patch_with(from_document(document), throw_on_extra_keys=False)


class SQLCombinerStore(CombinerStore, SQLStore[CombinerDTO, CombinerModel]):
    def __init__(self, Session):
        super().__init__(Session, CombinerModel)

    def get_by_name(self, name: str) -> CombinerDTO:
        with self.Session() as session:
            entity = session.query(CombinerModel).filter(CombinerModel.name == name).first()
            if entity is None:
                return None
            return self._dto_from_orm_model(entity)

    def _update_orm_model_from_dto(self, entity: CombinerModel, item: CombinerDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("combiner_id", None)
        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: CombinerModel) -> CombinerDTO:
        orm_dict = from_orm_model(item, CombinerModel)
        orm_dict["combiner_id"] = orm_dict.pop("id")
        return CombinerDTO().populate_with(orm_dict)



import click
import requests

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_api_url, get_response, get_token, print_response


@main.group("model")
@click.pass_context
def model_cmd(ctx):
    """:param ctx:"""
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-s", "--session_id", required=False, help="models in session with given session id")
@click.option("--n_max", required=False, help="Number of items to list")
@model_cmd.command("list")
@click.pass_context
def list_models(ctx, protocol: str, host: str, port: str, token: str = None, session_id: str = None, n_max: int = None):
    """List models."""
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    if session_id:
        response = get_response(
            protocol=protocol, host=host, port=port, endpoint=f"models/?session_id={session_id}/", token=token, headers=headers, usr_api=False, usr_token=False
        )
    else:
        response = get_response(protocol=protocol, host=host, port=port, endpoint="models/", token=token, headers=headers, usr_api=False, usr_token=False)
    print_response(response, "models", None)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-id", "--id", required=True, help="Model ID")
@model_cmd.command("get")
@click.pass_context
def get_model(ctx, protocol: str, host: str, port: str, token: str = None, id: str = None):
    """Get model by id."""
    response = get_response(protocol=protocol, host=host, port=port, endpoint=f"models/{id}", token=token, headers={}, usr_api=False, usr_token=False)
    print_response(response, "model", id)


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("-f", "--file", required=True, help="Path to the model file")
@model_cmd.command("set-active")
@click.pass_context
def set_active_model(ctx, protocol: str, host: str, port: str, token: str, file: str):
    """Set the initial model and upload to model repository."""
    headers = {}
    _token = get_token(token=token, usr_token=False)
    if _token:
        headers = {"Authorization": _token}

    if file.endswith(".npz"):
        helper = "numpyhelper"
    elif file.endswith(".bin"):
        helper = "binaryhelper"
    else:
        click.secho("Unsupported file type. Only .npz and .bin files are supported.", fg="red")
        return

    try:
        # Set the active helper
        url = get_api_url(protocol, host, port, "helpers/active", usr_api=False)
        response = requests.put(url, json={"helper": helper}, headers=headers, verify=False)
        response.raise_for_status()

        # Upload the model file
        url = get_api_url(protocol, host, port, "models/", usr_api=False)
        with open(file, "rb") as model_file:
            response = requests.post(url, files={"file": model_file}, data={"helper": helper}, headers=headers, verify=False)
            response.raise_for_status()

        click.secho("Model set as active and uploaded successfully.", fg="green")
    except requests.exceptions.RequestException as e:
        click.secho(f"Failed to set active model: {e}", fg="red")



from typing import Dict

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.validation import ValidationDTO
from fedn.network.storage.statestore.stores.sql.shared import ValidationModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class ValidationStore(Store[ValidationDTO]):
    pass


class MongoDBValidationStore(ValidationStore, MongoDBStore[ValidationDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "validation_id")

    def _document_from_dto(self, item: ValidationDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> ValidationDTO:
        dto_dict = from_document(document)

        return ValidationDTO().patch_with(dto_dict, throw_on_extra_keys=False, verify=True)


def _translate_key_sql(key: str) -> str:
    if key == "_id":
        key = "id"
    elif key == "sender.name":
        key = "sender_name"
    elif key == "sender.role":
        key = "sender_role"
    elif key == "receiver.name":
        key = "receiver_name"
    elif key == "receiver.role":
        key = "receiver_role"

    return key


class SQLValidationStore(ValidationStore, SQLStore[ValidationDTO, ValidationModel]):
    def __init__(self, Session):
        super().__init__(Session, ValidationModel)

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        kwargs = {_translate_key_sql(k): v for k, v in kwargs.items()}
        sort_key = _translate_key_sql(sort_key)
        return super().list(limit, skip, sort_key, sort_order, **kwargs)

    def count(self, **kwargs):
        kwargs = _translate_key_sql(kwargs)
        return super().count(**kwargs)

    def _update_orm_model_from_dto(self, entity: ValidationModel, item: ValidationDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("validation_id")

        sender: Dict = item_dict.pop("sender", {})
        item_dict["sender_name"] = sender.get("name")
        item_dict["sender_role"] = sender.get("role")

        receiver: Dict = item_dict.pop("receiver", {})
        item_dict["receiver_name"] = receiver.get("name")
        item_dict["receiver_role"] = receiver.get("role")

        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: ValidationModel) -> ValidationDTO:
        orm_dict = from_orm_model(item, ValidationModel)
        orm_dict["validation_id"] = orm_dict.pop("id")
        sender_name = orm_dict.pop("sender_name")
        sender_role = orm_dict.pop("sender_role")
        if sender_name is not None and sender_role is not None:
            orm_dict["sender"] = {"name": sender_name, "role": sender_role}
        reciever_name = orm_dict.pop("receiver_name")
        receiver_role = orm_dict.pop("receiver_role")
        if reciever_name is not None and receiver_role is not None:
            orm_dict["receiver"] = {"name": reciever_name, "role": receiver_role}

        return ValidationDTO().populate_with(orm_dict)



from abc import abstractmethod
from typing import Dict, List

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.round import RoundDTO
from fedn.network.storage.statestore.stores.sql.shared import (
    RoundCombinerDataModel,
    RoundCombinerModel,
    RoundConfigModel,
    RoundDataModel,
    RoundModel,
    from_orm_model,
)
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class RoundStore(Store[RoundDTO]):
    @abstractmethod
    def get_latest_round_id(self) -> int:
        pass


class MongoDBRoundStore(RoundStore, MongoDBStore[RoundDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "round_id")

    def update(self, item: RoundDTO) -> RoundDTO:
        return self.mongo_update(item)

    def get_latest_round_id(self) -> int:
        obj = self.database[self.collection].find_one(sort=[("_id", pymongo.DESCENDING)])
        if obj:
            return int(obj["round_id"])
        else:
            return 0

    def _document_from_dto(self, item: RoundDTO) -> Dict:
        return item.to_db(exclude_unset=False)

    def _dto_from_document(self, document: Dict) -> RoundDTO:
        return RoundDTO().patch_with(from_document(document), throw_on_extra_keys=False)


class SQLRoundStore(RoundStore, SQLStore[RoundDTO, RoundModel]):
    def __init__(self, Session):
        super().__init__(Session, RoundModel)

    def update(self, item: RoundDTO) -> RoundDTO:
        return self.sql_update(item)

    def get_latest_round_id(self) -> int:
        rounds = self.list(limit=1, skip=0, sort_key="round_id", sort_order=pymongo.DESCENDING)
        if any(rounds):
            return int(rounds[0].round_id)
        else:
            return 0

    def _update_orm_model_from_dto(self, entity: RoundModel, item: RoundDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("round_id", None)

        entity.id = item_dict["id"]
        entity.status = item_dict["status"]

        if "round_config" in item_dict and item_dict["round_config"] is not None:
            round_config = item_dict.pop("round_config")
            if entity.round_config is None:
                entity.round_config = RoundConfigModel(**round_config)
            else:
                self._patch_model(entity.round_config, round_config)
        else:
            entity.round_config = None

        if "round_data" in item_dict and item_dict["round_data"] is not None:
            round_data: Dict = item_dict.pop("round_data")
            reduce: Dict = round_data.pop("reduce", {})
            round_data["reduce_time_aggregate_model"] = reduce.pop("time_aggregate_model", None)
            round_data["reduce_time_fetch_model"] = reduce.pop("time_fetch_model", None)
            round_data["reduce_time_load_model"] = reduce.pop("time_load_model", None)
            if entity.round_data is None:
                entity.round_data = RoundDataModel(**round_data)
            else:
                self._patch_model(entity.round_data, round_data)
        else:
            entity.round_data = None

        round_combiners: List[Dict] = item_dict.pop("combiners")
        if round_combiners == self._combiners_to_dict_list(entity.combiners):
            # No changes do nothing
            pass
        else:
            entity.combiners.clear()
            for combiner in round_combiners:
                combiner_round_config: Dict = combiner.pop("config")
                combiner["config_job_id"] = combiner_round_config.pop("_job_id", None)

                combiner_round_config_model = RoundConfigModel(**combiner_round_config)

                combiner_data: Dict = combiner.pop("data")
                aggregation_time = combiner_data.pop("aggregation_time")
                combiner_data_model = RoundCombinerDataModel(**combiner_data)
                combiner_data_model.aggregation_time_nr_aggregated_models = aggregation_time["nr_aggregated_models"]
                combiner_data_model.aggregation_time_time_model_aggregation = aggregation_time["time_model_aggregation"]
                combiner_data_model.aggregation_time_time_model_load = aggregation_time["time_model_load"]

                round_combiner = RoundCombinerModel(**combiner)
                round_combiner.round_config = combiner_round_config_model
                round_combiner.data = combiner_data_model
                entity.combiners.append(round_combiner)

        return entity

    def _dto_from_orm_model(self, item: RoundModel) -> RoundDTO:
        orm_dict = from_orm_model(item, RoundModel)
        if item.round_config is not None:
            orm_dict["round_config"] = from_orm_model(item.round_config, RoundConfigModel)
            del orm_dict["round_config"]["id"]
            del orm_dict["round_config"]["committed_at"]

        if item.combiners is not None:
            orm_dict["combiners"] = self._combiners_to_dict_list(item.combiners)

        if item.round_data is not None:
            orm_dict["round_data"] = from_orm_model(item.round_data, RoundDataModel)
            orm_dict["round_data"]["reduce"] = {
                "time_aggregate_model": orm_dict["round_data"].pop("reduce_time_aggregate_model"),
                "time_fetch_model": orm_dict["round_data"].pop("reduce_time_fetch_model"),
                "time_load_model": orm_dict["round_data"].pop("reduce_time_load_model"),
            }
            del orm_dict["round_data"]["id"]
            del orm_dict["round_data"]["committed_at"]

        orm_dict["round_id"] = orm_dict.pop("id")
        del orm_dict["round_config_id"]
        del orm_dict["round_data_id"]

        return RoundDTO().populate_with(orm_dict)

    def _combiners_to_dict_list(self, combiners: List[RoundCombinerModel]):
        dict_combiners = []
        for c in combiners:
            c_dict = from_orm_model(c, RoundCombinerModel)
            if c.data is not None:
                c_dict["data"] = from_orm_model(c.data, RoundCombinerDataModel)
                c_dict["data"]["aggregation_time"] = {
                    "nr_aggregated_models": c_dict["data"].pop("aggregation_time_nr_aggregated_models"),
                    "time_model_aggregation": c_dict["data"].pop("aggregation_time_time_model_aggregation"),
                    "time_model_load": c_dict["data"].pop("aggregation_time_time_model_load"),
                }
                del c_dict["data"]["id"]
                del c_dict["data"]["committed_at"]

            if c.round_config is not None:
                c_dict["config"] = from_orm_model(c.round_config, RoundConfigModel)
                c_dict["config"]["_job_id"] = c_dict.pop("config_job_id")
                del c_dict["config"]["id"]
                del c_dict["config"]["committed_at"]

            del c_dict["id"]
            del c_dict["committed_at"]
            del c_dict["round_config_id"]
            del c_dict["data_id"]
            del c_dict["parent_round_id"]

            dict_combiners.append(c_dict)
        return dict_combiners

    def _patch_model(self, model, item_dict: Dict):
        for key, value in item_dict.items():
            setattr(model, key, value)
        return model



from abc import abstractmethod
from typing import Dict, List

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import select
from sqlalchemy.orm import aliased

from fedn.network.storage.statestore.stores.dto import ModelDTO
from fedn.network.storage.statestore.stores.shared import EntityNotFound
from fedn.network.storage.statestore.stores.sql.shared import ModelModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class ModelStore(Store[ModelDTO]):
    @abstractmethod
    def list_descendants(self, id: str, limit: int) -> List[ModelDTO]:
        """List descendants
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        return: A list of entities
        """
        pass

    @abstractmethod
    def list_ancestors(self, id: str, limit: int, include_self: bool = False, reverse: bool = False) -> List[ModelDTO]:
        """List ancestors
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        return: A list of entities
        """
        pass

    @abstractmethod
    def get_leaf_nodes(self) -> List[ModelDTO]:
        """List orphans
        return: A list of entities
        """
        pass

    @abstractmethod
    def get_active(self) -> str:
        """Get the active model
        return: The active model id (str)
        """
        pass

    @abstractmethod
    def set_active(self, id: str) -> bool:
        """Set the active model
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        return: True if successful
        """
        pass


class MongoDBModelStore(ModelStore, MongoDBStore[ModelDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "model")

    def get(self, id: str) -> ModelDTO:
        kwargs = {"key": "models", "model": id}
        document = self.database[self.collection].find_one(kwargs)
        if document is None:
            return None
        return self._dto_from_document(document)

    def update(self, item: ModelDTO) -> ModelDTO:
        item.check_validity()
        item_dict = self._document_from_dto(item)
        id = item_dict[self.primary_key]
        result = self.database[self.collection].update_one({self.primary_key: id, "key": "models"}, {"$set": item_dict})
        if result.matched_count == 1:
            document = self.database[self.collection].find_one({self.primary_key: id, "key": "models"})
            return self._dto_from_document(document)
        raise EntityNotFound(f"Entity with id {id} not found")

    def list(self, limit=0, skip=0, sort_key=None, sort_order=pymongo.DESCENDING, **kwargs):
        kwargs["key"] = "models"
        return super().list(limit, skip, sort_key, sort_order, **kwargs)

    def count(self, **kwargs) -> int:
        kwargs["key"] = "models"
        return super().count(**kwargs)

    def list_descendants(self, id: str, limit: int) -> List[ModelDTO]:
        kwargs = {"key": "models", "model": id}

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            return None

        current_model_id: str = model["model"]
        result: list = []

        for _ in range(limit):
            if current_model_id is None:
                break

            model: str = self.database[self.collection].find_one({"key": "models", "parent_model": current_model_id})

            if model is not None:
                formatted_model = self._dto_from_document(model)
                result.append(formatted_model)
                current_model_id = model["model_id"]
            else:
                break

        result.reverse()

        return result

    def list_ancestors(self, id: str, limit: int, include_self: bool = False, reverse: bool = False) -> List[ModelDTO]:
        kwargs = {"key": "models", "model": id}

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            return None

        current_model_id: str = model["parent_model"]
        result: list = []

        if include_self:
            formatted_model = self._dto_from_document(model)
            result.append(formatted_model)

        for _ in range(limit):
            if current_model_id is None:
                break

            model = self.database[self.collection].find_one({"key": "models", "model": current_model_id})

            if model is not None:
                formatted_model = self._dto_from_document(model)
                result.append(formatted_model)
                current_model_id = model["parent_model"]
            else:
                break

        if reverse:
            result.reverse()

        return result

    def get_leaf_nodes(self) -> List[ModelDTO]:
        parent_ids = self.database[self.collection].distinct("parent_model")

        kwargs = {"key": "models", "model": {"$nin": parent_ids}}
        leaf_nodes = list(self.database[self.collection].find(kwargs))

        result: list = []

        for model in leaf_nodes:
            formatted_model = self._dto_from_document(model)
            result.append(formatted_model)

        return result

    def get_active(self) -> str:
        active_model = self.database[self.collection].find_one({"key": "current_model"})

        if active_model is None:
            return None

        return active_model["model"]

    def set_active(self, id: str) -> bool:
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model = self.database[self.collection].find_one(kwargs)

        if model is None:
            return False

        self.database[self.collection].update_one({"key": "current_model"}, {"$set": {"model": model["model"]}}, upsert=True)

        return True

    def _document_from_dto(self, item: ModelDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        item_dict["model"] = item_dict.pop("model_id")
        item_dict["key"] = "models"
        return item_dict

    def _dto_from_document(self, document: Dict) -> ModelDTO:
        item_dict = from_document(document)
        item_dict["model_id"] = item_dict.pop("model")
        del item_dict["key"]
        return ModelDTO().patch_with(item_dict, throw_on_extra_keys=False)


class SQLModelStore(ModelStore, SQLStore[ModelDTO, ModelModel]):
    def __init__(self, Session):
        super().__init__(Session, ModelModel)

    def update(self, item: ModelDTO) -> ModelDTO:
        return self.sql_update(item)

    def list_descendants(self, id: str, limit: int):
        with self.Session() as session:
            # Define the recursive CTE
            descendant = aliased(ModelModel)  # Alias for recursion
            cte = select(ModelModel).where(ModelModel.parent_model == id).cte(name="descendant_cte", recursive=True)
            cte = cte.union_all(select(descendant).where(descendant.parent_model == cte.c.id))

            # Final query with optional limit
            query = select(cte)
            if limit is not None:
                query = query.limit(limit)

            # Execute the query
            items = session.execute(query).fetchall()

            # Return the list of descendants
            result = []
            for item in items:
                result.append(self._dto_from_orm_model(item))

            return result

    def list_ancestors(self, id: str, limit: int, include_self=False, reverse=False):
        with self.Session() as session:
            # Define the recursive CTE
            ancestor = aliased(ModelModel)  # Alias for recursion
            cte = select(ModelModel).where(ModelModel.id == id).cte(name="ancestor_cte", recursive=True)
            cte = cte.union_all(select(ancestor).where(ancestor.id == cte.c.parent_model))

            # Final query with optional limit
            query = select(cte)
            if limit is not None:
                query = query.limit(limit)

            # Execute the query
            items = session.execute(query).fetchall()

            # Return the list of ancestors
            result = []
            for item in items:
                result.append(self._dto_from_orm_model(item))

            return result

    def get_leaf_nodes(self) -> List[ModelDTO]:
        with self.Session() as session:
            # Define the recursive CTE
            child_alias = aliased(ModelModel)  # Alias for recursion
            query = session.query(ModelModel).outerjoin(child_alias, ModelModel.id == child_alias.parent_id).filter(child_alias.id.is_(None))
            items = session.execute(query).fetchall()

            # Return the list of ancestors
            result = []
            for item in items:
                result.append(self._dto_from_orm_model(item))

            return result

    def get_active(self) -> str:
        with self.Session() as session:
            active_stmt = select(ModelModel).where(ModelModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                return active_item.id
            return None

    def set_active(self, id: str) -> bool:
        with self.Session() as session:
            stmt = select(ModelModel).where(ModelModel.id == id)
            item = session.scalars(stmt).first()
            if item is None:
                return False

            active_stmt = select(ModelModel).where(ModelModel.active)
            active_items = session.scalars(active_stmt).all()
            for item in active_items:
                item.active = False
            item.active = True
            session.commit()
        return True

    def _update_orm_model_from_dto(self, entity: ModelModel, item: ModelDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("model_id", None)
        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: ModelModel) -> ModelDTO:
        orm_dict = from_orm_model(item, ModelModel)
        orm_dict["model_id"] = orm_dict.pop("id")
        del orm_dict["active"]
        return ModelDTO().populate_with(orm_dict)



"""Helper function to test if your server functions implementation runs correctly."""

import inspect
import json
from typing import Dict, List

import numpy as np

import fedn.network.grpc.fedn_pb2 as fedn
from fedn.network.combiner.hooks.hooks import FunctionServiceServicer
from fedn.network.combiner.hooks.serverfunctionsbase import ServerFunctionsBase
from fedn.network.combiner.modelservice import bytesIO_request_generator, model_as_bytesIO


def test_server_functions(server_functions: ServerFunctionsBase, parameters_np: List[np.ndarray], client_metadata: Dict, rounds=5, num_clients=20):
    """Test if your functionalities are working on your server functions implementation.
    :param server_functions: An implementation of ServerFunctionsBase.
    :type server_functions: ServerFunctionsBase
    :param parameters: Model parameters in a list of numpy arrays.
    :type parameters: List[np.ndarray]
    """
    function_service = FunctionServiceServicer()
    function_code = inspect.getsource(server_functions)
    for i in range(rounds):
        print(f"Simulating server round: {i+1}")
        # see output from provided functions call
        request = fedn.ProvidedFunctionsRequest(function_code=function_code)
        _ = function_service.HandleProvidedFunctions(request, "")
        # see output from client selection request
        fake_clients = [str(j) for j in range(num_clients)]
        request = fedn.ClientSelectionRequest(client_ids=json.dumps(fake_clients))
        response = function_service.HandleClientSelection(request, "")
        selected_clients = json.loads(response.client_ids)
        # see output from client config request
        bytesio_model = model_as_bytesIO(parameters_np)
        request_function = fedn.ClientConfigRequest
        args = {}
        gen = bytesIO_request_generator(mdl=bytesio_model, request_function=request_function, args=args)
        function_service.HandleClientConfig(gen, "")
        # see output from aggregate request
        request_function = fedn.StoreModelRequest
        args = {"id": "global_model"}
        bytesio_model = model_as_bytesIO(parameters_np)
        gen = bytesIO_request_generator(mdl=bytesio_model, request_function=request_function, args=args)
        function_service.HandleStoreModel(gen, "")
        for k in range(len(selected_clients)):
            # send metadata
            client_id = selected_clients[k]
            request = fedn.ClientMetaRequest(metadata=json.dumps(client_metadata), client_id=client_id)
            function_service.HandleMetadata(request, "")
            request_function = fedn.StoreModelRequest
            args = {"id": client_id}
            bytesio_model = model_as_bytesIO(parameters_np)
            gen = bytesIO_request_generator(mdl=bytesio_model, request_function=request_function, args=args)
            function_service.HandleStoreModel(gen, "")
        request = fedn.AggregationRequest(aggregate="aggregate")
        response_generator = function_service.HandleAggregation(request, "")
        for response in response_generator:
            pass
        print(f"Server round {i+1} completed")



import datetime
import itertools
import uuid

import pymongo
import pytest

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto import ModelDTO


@pytest.fixture
def test_models():
    return [ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name1"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name2"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name3"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name4"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model", session_id=None, name="test_name5"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model2", session_id=None, name="test_name6"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model2", session_id=None, name="test_name7"),
            ModelDTO(model_id=str(uuid.uuid4()), parent_model="test_parent_model2", session_id=None, name="test_name8")]


@pytest.fixture
def test_model():
    return ModelDTO(name="model_name", parent_model=None, session_id=None)


@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_models):
    for c in test_models:
        mongo_connection.model_store.add(c)
        postgres_connection.model_store.add(c)
        sql_connection.model_store.add(c)
        

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for m in test_models:
        mongo_connection.model_store.delete(m.model_id)
        postgres_connection.model_store.delete(m.model_id)
        sql_connection.model_store.delete(m.model_id)
        



@pytest.fixture
def options():
    sorting_keys = (None, 
                    "name",
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"parent_model":"test_parent_model2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestModelStore:

    def test_add_update_delete(self, db_connection:DatabaseConnection, test_model:ModelDTO):
        test_model.check_validity()

        # Add a model and check that we get the added model back
        read_model1 = db_connection.model_store.add(test_model)
        assert isinstance(read_model1.model_id, str)
        read_model1_dict = read_model1.to_dict()
        model_id = read_model1.model_id
        del read_model1_dict["model_id"]
        del read_model1_dict["model"]
        del read_model1_dict["committed_at"]


        input_dict = test_model.to_dict()
        del input_dict["model_id"]
        del input_dict["model"]
        del input_dict["committed_at"]

        assert read_model1_dict == input_dict

        # Assert we get the same model back
        read_model2 = db_connection.model_store.get(model_id)
        assert read_model2 is not None
        assert read_model2.to_dict() == read_model1.to_dict()
        
        # Update the model and check that we get the updated model back
        read_model2.name = "new_name"         
        read_model3 = db_connection.model_store.update(read_model2)
        assert read_model3.name == "new_name"

        # Assert we get the same model back
        read_model4 = db_connection.model_store.get(model_id)
        assert read_model4 is not None
        assert read_model3.to_dict() == read_model4.to_dict()


        # Delete the model and check that it is deleted
        success = db_connection.model_store.delete(model_id)
        assert success == True

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options: 
                gathered_models1 = db_1.model_store.list(*opt, **kwargs)
                count1 = db_1.model_store.count(**kwargs)

                gathered_models2 = db_2.model_store.list(*opt, **kwargs)
                count2 = db_2.model_store.count(**kwargs)

                assert count1 == count2
                assert len(gathered_models1) == len(gathered_models2)

                for i in range(len(gathered_models1)):
                    assert gathered_models1[i].model_id == gathered_models2[i].model_id
                    



import os
import time

import docker

CONTAINER_NAME="postgres-test-db"
""" 
    An alternative to this is to create the container using
    githiub action and then use the container to create the client
"""
def start_postgres_container():
    client = docker.from_env()        
    try:
        container = client.containers.get(CONTAINER_NAME)
    except docker.errors.NotFound:
        container = client.containers.run(
            "postgres:15",
            detach=True,
            ports={"5432/tcp": None}, # Let Docker choose an available port
            name=CONTAINER_NAME,
            environment={
                "POSTGRES_USER": os.environ.get("UNITTEST_DBUSER", "_"),
                "POSTGRES_PASSWORD": os.environ.get("UNITTEST_DBPASS", "_"),
                "POSTGRES_DB": "fedn_db"
            }
        )
    time.sleep(1)
    start = time.time()
    while time.time() - start < 10:
        try:
            container.reload()
            port = int(container.attrs['NetworkSettings']['Ports']['5432/tcp'][0]['HostPort'])
            break
        except:
            time.sleep(1)
    else:
        raise Exception("Could not start Postgres container")

    time.sleep(3)
    
    return container, port

def stop_postgres_container():
    client = docker.from_env()
    container = client.containers.get(CONTAINER_NAME)
    container.stop()
    container.remove(v=True, force=True)
    try:
        container.wait(condition="removed")
    except docker.errors.NotFound:
        pass





from datetime import datetime
from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, PrimaryID


class CombinerDTO(BaseDTO):
    """Client data transfer object."""

    combiner_id: Optional[str] = PrimaryID(None)
    name: str = Field(None)
    address: str = Field(None)
    fqdn: str = Field(None)
    ip: str = Field(None)
    parent: dict = Field(None)
    port: int = Field(None)
    updated_at: datetime = Field(None)



"""The aggregator package is responsible for aggregating models from multiple clients. It's called both in :class:`fedn.network.combiner.Combiner` and :class:`fedn.network.controller.Controller`
to aggregate models from clients.
"""
# ruff: noqa: E501



from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, DictDTO, Field, PrimaryID, validator
from fedn.network.storage.statestore.stores.shared import ValidationError


class SessionConfigDTO(DictDTO):
    """SessionConfig data transfer object."""

    aggregator: str = Field(None)
    aggregator_kwargs: Optional[str] = Field(None)
    round_timeout: int = Field(None)
    buffer_size: int = Field(None)
    rounds: Optional[int] = Field(None)
    delete_models_storage: bool = Field(None)
    clients_required: int = Field(None)
    requested_clients: Optional[int] = Field(None)
    validate: bool = Field(None)
    helper_type: str = Field(None)
    model_id: str = Field(None)
    server_functions: Optional[str] = Field(None)

    @validator
    def validate_config(self):
        if not self.aggregator:
            raise ValidationError("aggregator", "Aggregator is required")

        if not self.round_timeout:
            raise ValidationError("round_timeout", "Round timeout is required")

        if not self.buffer_size:
            raise ValidationError("buffer_size", "Buffer size is required")

        if not self.model_id:
            raise ValidationError("model_id", "Model ID is required")

        if not self.delete_models_storage:
            raise ValidationError("delete_models_storage", "Delete models storage is required")

        if not self.clients_required:
            raise ValidationError("clients_required", "Clients required is required")

        if not self.validate:
            raise ValidationError("validate", "Validate is required")

        if not self.helper_type:
            raise ValidationError("helper_type", "Helper type is required")


class SessionDTO(BaseDTO):
    """Session data transfer object."""

    session_id: str = PrimaryID(None)
    name: str = Field(None)
    status: Optional[str] = Field("Created")
    session_config: SessionConfigDTO = Field(SessionConfigDTO())

    @validator
    def validate(self):
        if not self.name:
            raise ValidationError("name", "Name is required")



from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, PrimaryID


class AnalyticDTO(BaseDTO):
    id: str = PrimaryID(None)
    sender_id: str = Field(None)
    sender_role: str = Field(None)
    memory_utilisation: float = Field(None)
    cpu_utilisation: float = Field(None)


# valid, msg = _validate_analytic(item_dict)

# def _validate_analytic(analytic: dict) -> Tuple[bool, str]:
#     if "sender_id" not in analytic:
#         return False, "sender_id is required"
#     if "sender_role" not in analytic or analytic["sender_role"] not in ["combiner", "client"]:
#         return False, "sender_role must be either 'combiner' or 'client'"
#     return analytic, ""



from typing import Dict

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto.status import StatusDTO
from fedn.network.storage.statestore.stores.sql.shared import StatusModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class StatusStore(Store[StatusDTO]):
    pass


class MongoDBStatusStore(StatusStore, MongoDBStore[StatusDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "status_id")

    def _document_from_dto(self, item: StatusDTO) -> Dict:
        item_dict = item.to_db(exclude_unset=False)
        return item_dict

    def _dto_from_document(self, document: Dict) -> StatusDTO:
        entity = from_document(document)

        if "logLevel" in entity:
            entity["log_level"] = entity["logLevel"]
        if "correlationId" in entity:
            entity["correlation_id"] = entity["correlationId"]
        if "sessionId" in entity:
            entity["session_id"] = entity["sessionId"]

        if "status_id" not in entity and "id" in entity:
            entity["status_id"] = entity["id"]

        return StatusDTO().patch_with(entity, throw_on_extra_keys=False)


def _translate_key_sql(key: str) -> str:
    if key == "_id":
        key = "id"
    elif key == "logLevel":
        key = "log_level"
    elif key == "sender.name":
        key = "sender_name"
    elif key == "sender.role":
        key = "sender_role"
    return key


class SQLStatusStore(StatusStore, SQLStore[StatusDTO, StatusModel]):
    def __init__(self, Session):
        super().__init__(Session, StatusModel)

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        kwargs = {_translate_key_sql(k): v for k, v in kwargs.items()}
        sort_key: str = _translate_key_sql(sort_key)
        return super().list(limit, skip, sort_key, sort_order, **kwargs)

    def count(self, **kwargs):
        kwargs = {_translate_key_sql(k): v for k, v in kwargs.items()}
        return super().count(**kwargs)

    def _update_orm_model_from_dto(self, entity: StatusModel, item: StatusDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("status_id", None)

        sender: Dict = item_dict.pop("sender", {})

        item_dict["sender_name"] = sender.get("name")
        item_dict["sender_role"] = sender.get("role")

        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: StatusModel) -> StatusDTO:
        orm_dict = from_orm_model(item, StatusModel)
        orm_dict["status_id"] = orm_dict.pop("id")
        orm_dict["sender"] = {
            "name": orm_dict.pop("sender_name"),
            "role": orm_dict.pop("sender_role"),
        }
        return StatusDTO().populate_with(orm_dict)



from flask import Blueprint, jsonify, request

from fedn.common.log_config import logger
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import metric_store
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers

bp = Blueprint("metric", __name__, url_prefix=f"/api/{api_version}/metrics")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_metrics():
    """Get metrics
    Retrieves a list of metrics based on the provided parameters.
    By specifying a parameter in the URL, you can filter the metrics based on that parameter,
    and the response will contain only the metrics that match the filter.
    ---
    tags:
        - Metrics
    parameters:
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: model_id
        in: query
        required: false
        type: string
        description: Model ID associated with the metric
      - name: model_step
        in: query
        required: false
        type: integer
        description: Model step associated with the metric
      - name: round_id
        in: query
        required: false
        type: string
        description: Round ID associated with the metric
      - name: session_id
        in: query
        required: false
        type: string
        description: Session ID associated with the metric
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of metrics to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of metrics to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the metrics by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the metrics in ('asc' or 'desc')
    definitions:
      Metric:
        type: object
        properties:
          metric_id:
            type: string
          key:
            type: string
          value:
            type: number
          timestamp:
            type: string
            format: date-time
          sender:
            type: object
            properties:
              name:
                type: string
              role:
                type: string
          model_id:
            type: string
          model_step:
            type: integer
          round_id:
            type: string
          session_id:
            type: string
    responses:
      200:
        description: A list of metrics and the total count.
        schema:
          type: object
          properties:
            count:
              type: integer
            result:
              type: array
              items:
                $ref: '#/definitions/Metric'
      500:
        description: An error occurred
        schema:
          type: object
          properties:
            message:
              type: string
    """
    try:
        limit, skip, sort_key, sort_order = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        result = metric_store.list(limit, skip, sort_key, sort_order, **kwargs)
        count = metric_store.count(**kwargs)
        response = {"count": count, "result": [item.to_dict() for item in result]}

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500


@jwt_auth_required(role="admin")
@bp.route("/list", methods=["POST"])
def list_metrics():
    """Retrieve a list of metrics based on the provided filters and pagination.

    This endpoint allows an admin user to fetch a list of metrics from the metric store.
    The results can be filtered, sorted, and paginated using the request headers and body.
    ---
    tags:
        - Metrics
    parameters:
    - name: limit
        in: header
        required: false
        type: integer
        description: The maximum number of metrics to retrieve
    - name: skip
        in: header
        required: false
        type: integer
        description: The number of metrics to skip
    - name: sort_key
        in: header
        required: false
        type: string
        description: The key to sort the metrics by
    - name: sort_order
        in: header
        required: false
        type: string
        description: The order to sort the metrics in ('asc' or 'desc')
    - name: filters
        in: body
        required: false
        schema:
        type: object
        additionalProperties:
            type: string
        description: Additional filters for querying metrics
    definitions:
    Metric:
        type: object
        properties:
        metric_id:
            type: string
        key:
            type: string
        value:
            type: number
        timestamp:
            type: string
            format: date-time
        sender:
            type: object
            properties:
            name:
                type: string
            role:
                type: string
        model_id:
            type: string
        model_step:
            type: integer
        round_id:
            type: string
        session_id:
            type: string
    responses:
    200:
        description: A list of metrics and the total count.
        schema:
        type: object
        properties:
            count:
            type: integer
            result:
            type: array
            items:
                $ref: '#/definitions/Metric'
    500:
        description: An error occurred
        schema:
        type: object
        properties:
            message:
            type: string
    """
    try:
        limit, skip, sort_key, sort_order = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        result = metric_store.list(limit, skip, sort_key, sort_order, **kwargs)
        count = metric_store.count(**kwargs)
        response = {"count": count, "result": [item.to_dict() for item in result]}

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500



"""DTOs for the StateStore."""

from fedn.network.storage.statestore.stores.dto.client import ClientDTO
from fedn.network.storage.statestore.stores.dto.combiner import CombinerDTO
from fedn.network.storage.statestore.stores.dto.model import ModelDTO
from fedn.network.storage.statestore.stores.dto.package import PackageDTO
from fedn.network.storage.statestore.stores.dto.prediction import NodeDTO, PredictionDTO
from fedn.network.storage.statestore.stores.dto.session import SessionConfigDTO, SessionDTO

__all__ = ["ClientDTO", "ModelDTO", "SessionConfigDTO", "SessionDTO", "CombinerDTO", "PackageDTO", "PredictionDTO", "NodeDTO"]



"""The controller package is responsible for orchestrating the federated learning process. It's acts as a gRPC client and sends round config tasks
to the :class:`fedn.network.combiner.Combiner`.
"""
# ruff: noqa: E501



"""Module contains minimal funcitonality mirrored from Pydantic BaseModel."""

import copy
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union, get_args, get_origin

from fedn.network.storage.statestore.stores.shared import MissingFieldError

T = TypeVar("T")


class Field:
    """Field class for DTOs."""

    def __init__(self, default_value) -> None:
        """Initialize Field with default value."""
        self.default_value = default_value

    def __deepcopy__(self, memo):
        return self.__class__(copy.deepcopy(self.default_value))


class validator:  # noqa: N801
    """decorator class for validation methods of DTOs."""

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class DTO(ABC):
    @abstractmethod
    def has_modifications(self) -> bool:
        """Check if DTO has any modifications."""
        pass

    @abstractmethod
    def model_dump(self, exclude_unset: bool = False) -> Dict[str, Any]:
        """Dump DTO to dict."""
        pass

    def to_dict(self) -> Dict[str, Any]:
        """Return dict representation of DTO.

        This method is used to convert the DTO to a dictionary representation.
        """
        return self.model_dump(exclude_unset=False)

    def to_db(self, exclude_unset: bool = False) -> Dict[str, Any]:
        """Return dict representation of DTO for database storage."""
        return self.model_dump(exclude_unset=exclude_unset)

    @abstractmethod
    def check_validity(self):
        """Verify DTO and raise ValidationError a if required field is missing or invalid."""
        pass

    @abstractmethod
    def patch_with(self, object: "DTO", throw_on_extra_keys: bool = True, verify: bool = False) -> "DTO":
        """Patch DTO with object.

        This method is used to apply a dict/list/DTO to the model with the option to throw an error if extra keys are present.
        It operates recursively if an dict/list are nested and matching fields are of type DTO.
        If verify is set to True, the DTO will be verified after the patch.
        """
        pass

    @abstractmethod
    def clear_all_changes(self):
        """Reset DTO to initial values."""
        pass


class DictDTO(DTO):
    def __init__(self, **kwargs) -> None:
        """Initialize DTO.

        This method copies all class fields and patches the instance with the provided kwargs.
        """
        for k in self.get_all_fieldnames():
            v = getattr(self.__class__, k).default_value
            super().__setattr__(k, copy.deepcopy(v))
        self._modified_fields = set()
        self.patch_with(kwargs)

    def __setattr__(self, name: str, value):
        """Set attribute and store which fields are modified."""
        if not self._is_field(name):
            super().__setattr__(name, value)
            return

        if issubclass(self._get_field_type(name), DTO):
            if value is None:
                super().__setattr__(name, None)
            elif isinstance(value, self._get_field_type(name)):
                super().__setattr__(name, value)
            elif issubclass(self._get_field_type(name), DictDTO):
                if not isinstance(value, dict):
                    raise ValueError(f"Can not set key: {name} to type {value.__class__.__name__} in {self.__class__.__name__}")
                new_value: DictDTO = self._get_field_type(name)()
                new_value.patch_with(value)
                super().__setattr__(name, new_value)
            elif issubclass(self._get_field_type(name), ListDTO):
                if not isinstance(value, list):
                    raise ValueError(f"Can not set key: {name} to type {value.__class__.__name__} in {self.__class__.__name__}")
                new_value: ListDTO = ListDTO(self._get_list_type(name))
                new_value.patch_with(value)
                super().__setattr__(name, new_value)
            else:
                raise ValueError(f"Can not set key: {name} to type {value.__class__.__name__} in {self.__class__.__name__}")
        else:
            super().__setattr__(name, value)
        self._modified_fields.add(name)

    def clear_field(self, fieldname):
        """Clear a field of modified value."""
        v = getattr(self.__class__, fieldname).default_value
        super().__setattr__(fieldname, copy.deepcopy(v))
        self._modified_fields.discard(fieldname)

    def clear_all_changes(self):
        """Clear all modified fields."""
        for field_name in self.get_all_fieldnames():
            self.clear_field(field_name)

    def model_dump(self, exclude_unset: bool = False) -> Dict[str, Any]:
        result = {}
        for field_name in self.get_all_fieldnames():
            field_value = getattr(self, field_name)
            if exclude_unset and not self._is_field_modified(field_name):
                continue
            if isinstance(field_value, DTO):
                result[field_name] = field_value.model_dump(exclude_unset)
            else:
                result[field_name] = field_value
        return result

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(exclude_unset=False)

    def to_db(self, exclude_unset: bool = False) -> Dict[str, Any]:
        return self.model_dump(exclude_unset=exclude_unset)

    def patch_with(self, value_dict: Union[Dict[str, Any], "DictDTO"], throw_on_extra_keys: bool = True, verify: bool = False) -> "DictDTO":
        if isinstance(value_dict, DictDTO):
            value_dict = value_dict.model_dump(exclude_unset=True)

        for key, value in value_dict.items():
            if self._is_field(key):
                setattr(self, key, value)
            elif throw_on_extra_keys:
                raise ValueError(f"Invalid key: {key} for {self.__class__.__name__}")

        if verify:
            self.check_validity()

        return self

    def populate_with(self, value_dict: Union[Dict[str, Any], "DictDTO"], throw_on_extra_keys: bool = True) -> "DictDTO":
        """Populate DTO from value_dict.

        This method is used to apply a dictionary to the model and requires all non-optional/previously unset fields to be present.
        It operates recursively if value_dict contains nested dictionaries and matching fields are of type DTO.

        A DTO populated with this method is guaranteed to have all required fields set.
        """
        return self.patch_with(value_dict, throw_on_extra_keys, verify=True)

    def has_modifications(self):
        if bool(self._modified_fields):
            return True
        return any(isinstance(getattr(self, field_name), DTO) and getattr(self, field_name).has_modifications() for field_name in self.get_all_fieldnames())

    def check_validity(self, exclude_primary_id=False):
        for field_name in self.get_all_fieldnames():
            field_value = getattr(self, field_name)
            if isinstance(field_value, DTO):
                if not self._is_field_optional(field_name) or field_value.has_modifications():
                    field_value.check_validity()
            elif not self._is_field_optional(field_name) and not self._is_field_modified(field_name):
                if not (self._is_primary_id(field_name) and exclude_primary_id):
                    raise MissingFieldError(field_name, self.__class__.__name__)
        self._run_validators()

    def _run_validators(self):
        for field_name in self.get_all_validators():
            validator = getattr(self, field_name)
            validator(self)

    @classmethod
    def get_all_fieldnames(cls) -> List[str]:
        """Get all fields of the class and its superclasses."""
        keys = []
        for base in cls.__mro__:
            if hasattr(base, "__dict__"):
                for key in base.__dict__.keys():
                    if cls._is_field(key):
                        keys.append(key)
        return keys

    @classmethod
    def get_all_validators(cls) -> List[str]:
        """Get all validators of the class and its superclasses."""
        keys = []
        for base in cls.__mro__:
            if hasattr(base, "__dict__"):
                for key in base.__dict__.keys():
                    if cls._is_validator(key):
                        keys.append(key)
        return keys

    # Private methods
    def _is_field_modified(self, field_name):
        """Check if a field is modified."""
        if field_name in self._modified_fields:
            return True
        field_value = getattr(self, field_name)
        if isinstance(field_value, DTO):
            return field_value.has_modifications()
        return False

    @classmethod
    def _is_field(cls, field_name: str) -> bool:
        """Check if a attribute is a Field."""
        return hasattr(cls, field_name) and isinstance(getattr(cls, field_name), Field)

    @classmethod
    def _is_primary_id(cls, field_name: str) -> bool:
        """Check if a attribute is a PrimaryID."""
        return hasattr(cls, field_name) and isinstance(getattr(cls, field_name), PrimaryID)

    @classmethod
    def _is_validator(cls, field_name: str) -> bool:
        """Check if a attribute is a validator."""
        return hasattr(cls, field_name) and isinstance(getattr(cls, field_name), validator)

    @classmethod
    def _is_field_optional(cls, field_name: str) -> bool:
        """Check if a field is optional."""
        for base in cls.__mro__:
            if hasattr(base, "__annotations__") and field_name in base.__annotations__:
                if cls._is_field(field_name):
                    return _is_optional(base.__annotations__[field_name])
        raise AttributeError(f"Field {field_name} not found in {cls.__name__}")

    @classmethod
    def _get_field_type(cls, field_name):
        """Get the type of a field."""
        for base in cls.__mro__:
            if hasattr(base, "__annotations__") and field_name in base.__annotations__:
                return _get_type(base.__annotations__[field_name])
        raise AttributeError(f"Field {field_name} not found in {cls.__name__}")

    @classmethod
    def _get_list_type(cls, field_name):
        """Get the type of a field."""
        for base in cls.__mro__:
            if hasattr(base, "__annotations__") and field_name in base.__annotations__:
                return _get_list_type(base.__annotations__[field_name])
        raise AttributeError(f"Field {field_name} not found in {cls.__name__}")

    def __deepcopy__(self, memo):
        return self.__class__(**copy.deepcopy(self.model_dump(exclude_unset=True)))


class ListDTO(DTO, Generic[T]):
    """ListDTO for Data Transfer Objects."""

    def __init__(self, ListClass, *values) -> None:
        """Initialize ListDTO."""
        super().__init__()
        self._ListClass = ListClass
        self.items: List[T] = []
        self._modified = False
        if values:
            self.patch_with(values)

    def model_dump(self, exclude_unset=False):
        if issubclass(self._ListClass, DTO):
            return [item.model_dump(exclude_unset) for item in self.items]
        return self.items

    def patch_with(self, value_list: Union[List[T], "ListDTO"], throw_on_extra_keys: bool = True, verify: bool = False) -> "ListDTO":
        """Patch ListDTO with list."""
        if isinstance(value_list, ListDTO):
            value_list = value_list.model_dump(exclude_unset=True)

        self.items.clear()
        self._modified = True

        for i, item in enumerate(value_list):
            if issubclass(self._ListClass, DTO):
                dto = self._ListClass()
                dto.patch_with(item, throw_on_extra_keys)
                self.items.append(dto)
            else:
                self.items.append(item)

        if verify:
            self.check_validity()

        return self

    def check_validity(self):
        if issubclass(self._ListClass, DTO):
            for item in self.items:
                item.check_validity()

    def clear_all_changes(self):
        self.items.clear()
        self._modified = False

    def has_modifications(self) -> bool:
        return self._modified or any(isinstance(item, DTO) and item.has_modifications() for item in self.items)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, key) -> T:
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    def append(self, item: T):
        if issubclass(self._ListClass, DTO):
            item = self._ListClass().patch_with(item)
            self.items.append(item)
        else:
            self.items.append(item)
        self._modified = True

    def copy(self):
        return copy.deepcopy(self)

    def __deepcopy__(self, memo):
        return self.__class__(self._ListClass, *copy.deepcopy(self.model_dump(exclude_unset=True)))


class PrimaryID(Field):
    """PrimaryID field for DTOs."""

    pass


class BaseDTO(DictDTO):
    """BaseDTO for Data Transfer Objects."""

    committed_at: datetime = Field(None)

    @property
    def primary_id(self) -> str:
        """Get the id of the DTO."""
        for base in self.__class__.__mro__:
            if hasattr(base, "__dict__"):
                for key in base.__dict__.keys():
                    if isinstance(getattr(self.__class__, key), PrimaryID):
                        return getattr(self, key)
        raise AttributeError(f"{self.__class__.__name__} has no field of type PrimaryID")

    def _is_field_optional(self, key):
        return super()._is_field_optional(key) or key == "committed_at"


class NodeDTO(DictDTO):
    """Agent data transfer object."""

    name: str = Field(None)
    role: str = Field(None)
    client_id: Optional[str] = Field(None)


# Private functions


def _is_optional(annotation) -> bool:
    """Check if a type annotation is Optional."""
    return get_origin(annotation) is Union and type(None) in get_args(annotation)


def _get_type(annotation) -> type:
    """Get the type of a type annotation. If the annotation is Optional, return the inner type."""
    if get_origin(annotation) is Union:
        return get_args(annotation)[0]
    elif get_origin(annotation) is not None:
        return get_origin(annotation)
    return annotation


def _get_list_type(annotation) -> type:
    """Get the type of a type annotation. If the annotation is Optional, return the inner type."""
    if get_origin(annotation) is ListDTO:
        return get_args(annotation)[0]
    else:
        return None



from abc import abstractmethod
from typing import Dict, List

from pymongo.database import Database
from sqlalchemy import func, select

from fedn.network.storage.statestore.stores.dto import ClientDTO
from fedn.network.storage.statestore.stores.sql.shared import ClientModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class ClientStore(Store[ClientDTO]):
    """Client store interface."""

    @abstractmethod
    def connected_client_count(self, combiners: List[str]) -> List[ClientDTO]:
        """Count the number of connected clients for each combiner.

        :param combiners: list of combiners to get data for.
        :type combiners: list
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :return: list of combiner data.
        :rtype: list(ObjectId)
        """
        pass


class MongoDBClientStore(ClientStore, MongoDBStore[ClientDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "client_id")

    def update(self, item: ClientDTO) -> ClientDTO:
        return self.mongo_update(item)

    def connected_client_count(self, combiners: List[str]) -> List:
        try:
            pipeline = (
                [
                    {"$match": {"combiner": {"$in": combiners}, "status": "online"}},
                    {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                    {"$project": {"id": "$_id", "count": 1, "_id": 0}},
                ]
                if len(combiners) > 0
                else [
                    {"$match": {"status": "online"}},
                    {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                    {"$project": {"id": "$_id", "count": 1, "_id": 0}},
                ]
            )

            result = list(self.database[self.collection].aggregate(pipeline))
        except Exception:
            result = []

        return result

    def _dto_from_document(self, document: Dict) -> ClientDTO:
        return ClientDTO().patch_with(from_document(document), throw_on_extra_keys=False)

    def _document_from_dto(self, item: ClientDTO) -> Dict:
        return item.to_db(exclude_unset=False)


class SQLClientStore(ClientStore, SQLStore[ClientDTO, ClientModel]):
    def __init__(self, Session):
        super().__init__(Session, ClientModel)

    def update(self, item: ClientDTO) -> ClientDTO:
        return self.sql_update(item)

    def connected_client_count(self, combiners) -> List[Dict]:
        with self.Session() as session:
            stmt = select(ClientModel.combiner, func.count(ClientModel.combiner)).group_by(ClientModel.combiner)
            if combiners:
                stmt = stmt.where(ClientModel.combiner.in_(combiners))

            items = session.execute(stmt).fetchall()

            result = []
            for item in items:
                result.append({"combiner": item[0], "count": item[1]})

            return result

    def _update_orm_model_from_dto(self, entity: ClientModel, item: ClientDTO) -> ClientModel:
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("client_id", None)
        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: ClientModel) -> ClientDTO:
        orm_dict = from_orm_model(item, ClientModel)
        orm_dict["client_id"] = orm_dict.pop("id")
        return ClientDTO().populate_with(orm_dict)



import datetime
import itertools
import uuid

import pymongo
import pytest

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.network.storage.statestore.stores.dto import ClientDTO


@pytest.fixture
def test_clients():
    #TODO: SQL versions does not support updated_at field
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    return [
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client1", combiner="test_combiner", ip="121.12.32.2", combiner_preferred="", status="test_status", last_seen=start_date - datetime.timedelta(days=1), package="local"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client2", combiner="test_combiner", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(days=3), package="local"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client3", combiner="test_combiner", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(days=4), package="local"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client5", combiner="test_combiner", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(seconds=5), package="remote"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client4", combiner="test_combiner2", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(weeks=1), package="remote"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client6", combiner="test_combiner2", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(hours=1), package="local"),
        ClientDTO(client_id=str(uuid.uuid4()), name="test_client7", combiner="test_combiner2", ip="121.12.32.22", combiner_preferred="", status="test_status2", last_seen=start_date - datetime.timedelta(minutes=1), package="remote"),
    ]

@pytest.fixture
def test_client():
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    c = ClientDTO(name = "name", combiner = "combiner", combiner_preferred = "combiner_preferred", ip = "ip", status = "status", last_seen = start_date, package = "package")
    return c

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_clients):
    for c in test_clients:
        mongo_connection.client_store.add(c)
        postgres_connection.client_store.add(c)
        sql_connection.client_store.add(c)

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    for c in test_clients:
        mongo_connection.client_store.delete(c.client_id)
        postgres_connection.client_store.delete(c.client_id)
        sql_connection.client_store.delete(c.client_id)



@pytest.fixture
def options():
    sorting_keys = (None,
                    "name",
                    "client_id",
                    "last_seen",
                    "ip", # None unique key 
                    "invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"ip":"121.12.32.22"}, {"combiner":"test_combiner", "status":"test_status"}, {"name":"test_client1", "status":"test_status2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))


class TestClientStore:

    def test_add_get_update_delete(self, db_connection:DatabaseConnection, test_client: ClientDTO): 
        test_client.check_validity()

        # Add a client and check that we get the added client back
        read_client1 = db_connection.client_store.add(test_client)
        assert isinstance(read_client1.client_id, str)
        read_client1_dict = read_client1.to_dict()
        client_id = read_client1_dict["client_id"]
        del read_client1_dict["client_id"]
        del read_client1_dict["committed_at"]

        test_client_dict = test_client.to_dict()
        del test_client_dict["client_id"]
        del test_client_dict["committed_at"]

        assert read_client1_dict == test_client_dict

        # Assert we get the same client back
        read_client2 = db_connection.client_store.get(client_id)
        assert read_client2 is not None
        assert read_client2.to_dict() == read_client1.to_dict()
        
        # Update the client and check that we get the updated client back
        read_client2.name = "new_name"         
        read_client3 = db_connection.client_store.update(read_client2)
        assert read_client3.name == "new_name"

        # Assert we get the same client back
        read_client4 = db_connection.client_store.get(client_id)
        assert read_client4 is not None
        assert read_client3.to_dict() == read_client4.to_dict()

        # Delete the client and check that it is deleted
        success = db_connection.client_store.delete(client_id)
        assert success == True

    def test_list_count(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:                        
                gathered_clients = db_1.client_store.list(*opt, **kwargs)
                gathered_clients2 = db_2.client_store.list(*opt, **kwargs)
                count = db_1.client_store.count(**kwargs)
                count2 = db_2.client_store.count(**kwargs)

                assert count == count2
                assert len(gathered_clients) == len(gathered_clients2)

                for i in range(len(gathered_clients)):
                    assert gathered_clients[i].client_id == gathered_clients2[i].client_id
                





    
    
    
    
    



from typing import Optional

from fedn.network.storage.statestore.stores.dto.shared import BaseDTO, Field, PrimaryID


class ModelDTO(BaseDTO):
    """Model data transfer object."""

    model_id: Optional[str] = PrimaryID(None)
    name: str = Field(None)
    parent_model: Optional[str] = Field(None)
    session_id: Optional[str] = Field(None)

    def to_dict(self):
        res = super().to_dict()
        # TODO: Remove this when we have migrated all model to model_id
        res["model"] = self.model_id
        return res

    def patch_with(self, value_dict, throw_on_extra_keys=True, verify=False):
        # TODO: Remove this when we have migrated all model to model_id
        if "model" in value_dict:
            if "model_id" in value_dict and value_dict["model_id"] != value_dict["model"]:
                raise ValueError("Cannot set both model and model_id in ModelDTO")
            value_dict["model_id"] = value_dict.pop("model")
        return super().patch_with(value_dict, throw_on_extra_keys)



from typing import Dict, List

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.dto import PredictionDTO
from fedn.network.storage.statestore.stores.sql.shared import PredictionModel, from_orm_model
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store, from_document


class PredictionStore(Store[PredictionDTO]):
    pass


class MongoDBPredictionStore(PredictionStore, MongoDBStore[PredictionDTO]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection, "prediction_id")

    def _document_from_dto(self, item: PredictionDTO) -> Dict:
        doc = item.to_db()
        return doc

    def _dto_from_document(self, document: Dict) -> PredictionDTO:
        item = from_document(document)

        pred = PredictionDTO()
        pred.sender.patch_with(item.pop("sender"), throw_on_extra_keys=False)
        pred.receiver.patch_with(item.pop("receiver"), throw_on_extra_keys=False)
        pred.patch_with(item, throw_on_extra_keys=False)
        return pred


def _translate_key_sql(key: str):
    if key == "sender.name":
        key = "sender_name"
    elif key == "sender.role":
        key = "sender_role"
    elif key == "receiver.name":
        key = "receiver_name"
    elif key == "receiver.role":
        key = "receiver_role"
    return key


class SQLPredictionStore(PredictionStore, SQLStore[PredictionDTO, PredictionModel]):
    def __init__(self, Session):
        super().__init__(Session, PredictionModel)

    def list(self, limit: int = 0, skip: int = 0, sort_key: str = None, sort_order=pymongo.DESCENDING, **kwargs) -> List[PredictionDTO]:
        kwargs = {_translate_key_sql(k): v for k, v in kwargs.items()}
        sort_key = _translate_key_sql(sort_key)
        return super().list(limit, skip, sort_key, sort_order, **kwargs)

    def count(self, **kwargs):
        kwargs = {_translate_key_sql(k): v for k, v in kwargs.items()}
        return super().count(**kwargs)

    def _update_orm_model_from_dto(self, entity: PredictionModel, item: PredictionDTO):
        item_dict = item.to_db(exclude_unset=False)
        item_dict["id"] = item_dict.pop("prediction_id", None)

        if "sender" in item_dict:
            sender: Dict = item_dict.pop("sender")
            item_dict["sender_name"] = sender.get("name")
            item_dict["sender_role"] = sender.get("role")
        if "receiver" in item_dict:
            receiver: Dict = item_dict.pop("receiver")
            item_dict["receiver_name"] = receiver.get("name")
            item_dict["receiver_role"] = receiver.get("role")

        for key, value in item_dict.items():
            setattr(entity, key, value)
        return entity

    def _dto_from_orm_model(self, item: PredictionModel) -> PredictionDTO:
        orm_dict = from_orm_model(item, PredictionModel)
        orm_dict["sender"] = {"name": orm_dict.pop("sender_name"), "role": orm_dict.pop("sender_role")}
        orm_dict["receiver"] = {"name": orm_dict.pop("receiver_name"), "role": orm_dict.pop("receiver_role")}
        orm_dict["prediction_id"] = orm_dict.pop("id")
        pred = PredictionDTO()
        pred.populate_with(orm_dict)
        return pred



import uuid
from datetime import datetime
from typing import Dict, List, Optional, Type

from sqlalchemy import JSON, ForeignKey, MetaData, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

constraint_naming_conventions = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


def from_orm_model(model, SQLModel: Type[DeclarativeBase]):
    result = {}
    for k in SQLModel.__table__.columns:
        result[k.name] = getattr(model, k.name)
    return result


class Base(DeclarativeBase):
    metadata = MetaData(naming_convention=constraint_naming_conventions)


class MyAbstractBase(Base):
    __abstract__ = True

    id: Mapped[str] = mapped_column(primary_key=True, default=lambda: str(uuid.uuid4()))
    committed_at: Mapped[datetime] = mapped_column(default=datetime.now)


class SessionConfigModel(MyAbstractBase):
    __tablename__ = "session_configs"

    aggregator: Mapped[str] = mapped_column(String(255))
    aggregator_kwargs: Mapped[Optional[str]]
    round_timeout: Mapped[int]
    buffer_size: Mapped[int]
    delete_models_storage: Mapped[bool]
    clients_required: Mapped[int]
    requested_clients: Mapped[Optional[int]]
    validate: Mapped[bool]
    helper_type: Mapped[str] = mapped_column(String(255))
    model_id: Mapped[str] = mapped_column(ForeignKey("models.id"))
    session: Mapped["SessionModel"] = relationship(back_populates="session_config")
    rounds: Mapped[int]
    server_functions: Mapped[Optional[str]] = mapped_column(String(255))


class SessionModel(MyAbstractBase):
    __tablename__ = "sessions"

    name: Mapped[Optional[str]] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    session_config_id: Mapped[str] = mapped_column(ForeignKey("session_configs.id"))
    session_config: Mapped["SessionConfigModel"] = relationship(back_populates="session", cascade="all, delete-orphan", single_parent=True)
    models: Mapped[List["ModelModel"]] = relationship(back_populates="session")


class ModelModel(MyAbstractBase):
    __tablename__ = "models"

    active: Mapped[bool] = mapped_column(default=False)
    parent_model: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[Optional[str]] = mapped_column(String(255))
    session_configs: Mapped[List["SessionConfigModel"]] = relationship()
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    session: Mapped[Optional["SessionModel"]] = relationship(back_populates="models")


class RoundConfigModel(MyAbstractBase):
    __tablename__ = "round_configs"

    aggregator: Mapped[str] = mapped_column(String(255))
    aggregator_kwargs: Mapped[Optional[str]]
    round_timeout: Mapped[int]
    buffer_size: Mapped[int]
    delete_models_storage: Mapped[bool]
    clients_required: Mapped[int]
    requested_clients: Mapped[Optional[int]]
    validate: Mapped[bool]
    helper_type: Mapped[str] = mapped_column(String(255))
    server_functions: Mapped[Optional[str]]
    model_id: Mapped[Optional[str]] = mapped_column(ForeignKey("models.id"))
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    task: Mapped[str] = mapped_column(String(255))
    round_id: Mapped[str]
    rounds: Mapped[int]
    client_settings: Mapped[Optional[Dict]] = mapped_column(JSON)


class RoundCombinerDataModel(MyAbstractBase):
    __tablename__ = "round_combiner_data"

    aggregation_time_nr_aggregated_models: Mapped[Optional[int]]
    aggregation_time_time_model_aggregation: Mapped[Optional[float]]
    aggregation_time_time_model_load: Mapped[Optional[float]]
    nr_expected_updates: Mapped[Optional[int]]
    nr_required_updates: Mapped[Optional[int]]
    time_combination: Mapped[Optional[float]]
    timeout: Mapped[Optional[float]]
    round_combiner: Mapped["RoundCombinerModel"] = relationship(back_populates="data")


class RoundCombinerModel(MyAbstractBase):
    __tablename__ = "round_combiners"

    model_id: Mapped[str]
    name: Mapped[str] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    time_exec_training: Mapped[float]

    round_config_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_configs.id"))
    round_config: Mapped[Optional["RoundConfigModel"]] = relationship(cascade="all, delete-orphan", single_parent=True)

    data_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_combiner_data.id"))
    data: Mapped[Optional["RoundCombinerDataModel"]] = relationship(back_populates="round_combiner", cascade="all, delete")

    parent_round_id: Mapped[str] = mapped_column(ForeignKey("rounds.id"))
    round: Mapped["RoundModel"] = relationship(back_populates="combiners")

    config_job_id: Mapped[Optional[str]]
    round_id: Mapped[Optional[str]]


class RoundDataModel(MyAbstractBase):
    __tablename__ = "round_data"

    time_commit: Mapped[Optional[float]]
    reduce_time_aggregate_model: Mapped[Optional[float]]
    reduce_time_fetch_model: Mapped[Optional[float]]
    reduce_time_load_model: Mapped[Optional[float]]
    round: Mapped["RoundModel"] = relationship(back_populates="round_data")


class RoundModel(MyAbstractBase):
    __tablename__ = "rounds"
    status: Mapped[str] = mapped_column()

    round_config_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_configs.id", ondelete="cascade"))
    round_config: Mapped[Optional["RoundConfigModel"]] = relationship(cascade="all, delete-orphan", single_parent=True)
    round_data_id: Mapped[Optional[str]] = mapped_column(ForeignKey("round_data.id", ondelete="cascade"))
    round_data: Mapped[Optional["RoundDataModel"]] = relationship(back_populates="round", cascade="all, delete")
    combiners: Mapped[List["RoundCombinerModel"]] = relationship(back_populates="round", cascade="all, delete-orphan")


class ClientModel(MyAbstractBase):
    __tablename__ = "clients"

    combiner: Mapped[str] = mapped_column(String(255))
    combiner_preferred: Mapped[str] = mapped_column(String(255))
    ip: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    package: Mapped[Optional[str]] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    last_seen: Mapped[datetime] = mapped_column(default=datetime.now())


class PackageModel(MyAbstractBase):
    __tablename__ = "packages"

    active: Mapped[bool] = mapped_column(default=False)
    description: Mapped[Optional[str]] = mapped_column(String(255))
    file_name: Mapped[str] = mapped_column(String(255))
    helper: Mapped[str] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    storage_file_name: Mapped[str] = mapped_column(String(255))


class CombinerModel(MyAbstractBase):
    __tablename__ = "combiners"

    address: Mapped[str] = mapped_column(String(255))
    fqdn: Mapped[Optional[str]] = mapped_column(String(255))
    ip: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    parent: Mapped[Optional[str]] = mapped_column(String(255))
    port: Mapped[int]
    updated_at: Mapped[datetime] = mapped_column(default=datetime.now())


class PredictionModel(MyAbstractBase):
    __tablename__ = "predictions"

    correlation_id: Mapped[str]
    data: Mapped[Optional[str]]
    model_id: Mapped[Optional[str]] = mapped_column(ForeignKey("models.id"))
    receiver_name: Mapped[Optional[str]] = mapped_column(String(255))
    receiver_role: Mapped[Optional[str]] = mapped_column(String(255))
    sender_name: Mapped[Optional[str]] = mapped_column(String(255))
    sender_role: Mapped[Optional[str]] = mapped_column(String(255))
    meta: Mapped[Optional[str]] = mapped_column(String(255))
    timestamp: Mapped[str] = mapped_column(String(255))


class StatusModel(MyAbstractBase):
    __tablename__ = "statuses"

    log_level: Mapped[str] = mapped_column(String(255))
    sender_name: Mapped[Optional[str]] = mapped_column(String(255))
    sender_role: Mapped[Optional[str]] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    timestamp: Mapped[datetime]
    type: Mapped[str] = mapped_column(String(255))
    data: Mapped[Optional[str]]
    correlation_id: Mapped[Optional[str]]
    extra: Mapped[Optional[str]]
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))


class ValidationModel(MyAbstractBase):
    __tablename__ = "validations"

    correlation_id: Mapped[str]
    data: Mapped[Optional[str]]
    model_id: Mapped[Optional[str]] = mapped_column(ForeignKey("models.id"))
    meta: Mapped[Optional[str]] = mapped_column(String(255))
    receiver_name: Mapped[Optional[str]] = mapped_column(String(255))
    receiver_role: Mapped[Optional[str]] = mapped_column(String(255))
    sender_name: Mapped[Optional[str]] = mapped_column(String(255))
    sender_role: Mapped[Optional[str]] = mapped_column(String(255))
    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    timestamp: Mapped[datetime]


class MetricModel(MyAbstractBase):
    __tablename__ = "metrics"

    key: Mapped[str] = mapped_column(String(255))
    value: Mapped[float]

    # Client timestamp
    timestamp: Mapped[Optional[datetime]]

    sender_id: Mapped[str]
    sender_role: Mapped[str]

    model_id: Mapped[str] = mapped_column(ForeignKey("models.id"))
    step: Mapped[Optional[int]]

    session_id: Mapped[Optional[str]] = mapped_column(ForeignKey("sessions.id"))
    round_id: Mapped[Optional[str]] = mapped_column(ForeignKey("rounds.id"))
